from abc import ABC, abstractmethod
from typing import Generic, TypeVar

UID = TypeVar("UID")
GID = TypeVar("GID")


class PostDestination(ABC, Generic[UID, GID]):

    @property
    def adapter(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def user_id(self) -> UID:
        raise NotImplementedError()

    @property
    @abstractmethod
    def group_id(self) -> GID:
        raise NotImplementedError()


class PostDestinationFactory(ABC, Generic[UID, GID]):
    @abstractmethod
    def build(self, user_id: UID, group_id: GID) -> PostDestination:
        raise NotImplementedError()



from motor.motor_asyncio import AsyncIOMotorClient
from nonebot import logger, get_driver
from pymongo.errors import OperationFailure

from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.data.source.mongo.migration import MongoMigrationManager
from nonebot_plugin_pixivbot.global_context import context as context


@context.register_singleton()
class MongoDataSource:
    conf = context.require(Config)
    app_db_version = 2

    def __init__(self):
        self._client = None
        self._db = None

        get_driver().on_startup(self.initialize)
        get_driver().on_shutdown(self.finalize)

    @property
    def client(self):
        return self._client

    @property
    def db(self):
        return self._db

    async def get_db_version(self) -> int:
        version = await self.db["meta_info"].find_one({"key": "db_version"})
        if version is None:
            await self.db["meta_info"].insert_one({"key": "db_version", "value": 1})
            return 1
        else:
            return version["value"]

    async def set_db_version(self, db_version: int):
        await self.db["meta_info"].update_one({"key": "db_version"},
                                              {"$set": {
                                                  "value": db_version
                                              }},
                                              upsert=True)

    async def _ensure_index(self, coll_name: str, indexes: list[tuple[str, int]], **kwargs):
        try:
            await self.db[coll_name].create_index(indexes, **kwargs)
        except OperationFailure:
            logger.info(f"Index in {coll_name}: recreated")
            await self.db[coll_name].drop_index(indexes)
            await self.db[coll_name].create_index(indexes, **kwargs)

    async def _ensure_ttl_index(self, coll_name: str, expires_in: int):
        try:
            await self.db[coll_name].create_index([("update_time", 1)], expireAfterSeconds=expires_in)
        except OperationFailure:
            await self.db.command({
                "collMod": coll_name,
                "index": {
                    "keyPattern": {"update_time": 1},
                    "expireAfterSeconds": expires_in,
                }
            })
            logger.success(f"TTL Index in {coll_name}: expireAfterSeconds changed to {expires_in}")

    async def initialize(self):
        self._client = AsyncIOMotorClient(self.conf.pixiv_mongo_conn_url)
        self._db = self._client[self.conf.pixiv_mongo_database_name]

        # migrate
        db_version = await self.get_db_version()
        await context.require(MongoMigrationManager).perform_migration(self._db, db_version, self.app_db_version)
        await self.set_db_version(self.app_db_version)

        # ensure index
        await self._ensure_index('meta_info', [("key", 1)], unique=True)

        await self._ensure_index('pixiv_binding', [("adapter", 1), ("user_id", 1)], unique=True)

        await self._ensure_index('subscription', [("adapter", 1), ("user_id", 1), ("type", 1)])
        await self._ensure_index('subscription', [("adapter", 1), ("group_id", 1), ("type", 1)])

        await self._ensure_index('local_tags', [("name", 1)], unique=True)
        await self._ensure_index('local_tags', [("translated_name", 1)])

        await self._ensure_index('download_cache', [("illust_id", 1)], unique=True)
        await self._ensure_ttl_index('download_cache', self.conf.pixiv_download_cache_expires_in)

        await self._ensure_index('illust_detail_cache', [("illust.id", 1)], unique=True)
        await self._ensure_ttl_index('illust_detail_cache', self.conf.pixiv_illust_detail_cache_expires_in)

        await self._ensure_index('user_detail_cache', [("user.id", 1)], unique=True)
        await self._ensure_ttl_index('user_detail_cache', self.conf.pixiv_user_detail_cache_expires_in)

        await self._ensure_index('illust_ranking_cache', [("mode", 1)], unique=True)
        await self._ensure_ttl_index('illust_ranking_cache', self.conf.pixiv_illust_ranking_cache_expires_in)

        await self._ensure_index('search_illust_cache', [("word", 1)], unique=True)
        await self._ensure_ttl_index('search_illust_cache', self.conf.pixiv_search_illust_cache_expires_in)

        await self._ensure_index('search_user_cache', [("word", 1)], unique=True)
        await self._ensure_ttl_index('search_user_cache', self.conf.pixiv_search_user_cache_expires_in)

        await self._ensure_index('user_illusts_cache', [("user_id", 1)], unique=True)
        await self._ensure_ttl_index('user_illusts_cache', self.conf.pixiv_user_illusts_cache_expires_in)

        await self._ensure_index('user_bookmarks_cache', [("user_id", 1)], unique=True)
        await self._ensure_ttl_index('user_bookmarks_cache', self.conf.pixiv_user_bookmarks_cache_expires_in)

        await self._ensure_index('related_illusts_cache', [("original_illust_id", 1)], unique=True)
        await self._ensure_ttl_index('related_illusts_cache', self.conf.pixiv_related_illusts_cache_expires_in)

        await self._ensure_index('other_cache', [("type", 1)], unique=True)
        await self._ensure_ttl_index('other_cache', self.conf.pixiv_other_cache_expires_in)
        logger.success("ensure indexes completed")

    async def finalize(self):
        self._client.close()
        self._client = None
        self._db = None


__all__ = ("MongoDataSource",)



from typing import TypeVar, Generic, Sequence, Any

from nonebot_plugin_pixivbot.global_context import context as context
from nonebot_plugin_pixivbot.handler.cmd.command_handler import SubCommandHandler, CommandHandler
from nonebot_plugin_pixivbot.postman import PostDestination
from nonebot_plugin_pixivbot.service.pixiv_account_binder import PixivAccountBinder
from nonebot_plugin_pixivbot.utils.errors import BadRequestError

UID = TypeVar("UID")
GID = TypeVar("GID")


@context.require(CommandHandler).sub_command("bind")
class BindHandler(SubCommandHandler[UID, GID], Generic[UID, GID]):
    binder = context.require(PixivAccountBinder)

    @classmethod
    def type(cls) -> str:
        return "bind"

    @classmethod
    def enabled(cls) -> bool:
        return True

    def parse_args(self, args: Sequence[Any], post_dest: PostDestination[UID, GID]) -> dict:
        if len(args) < 1:
            raise BadRequestError()
        else:
            pixiv_user_id = int(args[0])
            return {"pixiv_user_id": pixiv_user_id}

    async def actual_handle(self, *, pixiv_user_id: int,
                            post_dest: PostDestination[UID, GID],
                            silently: bool = False):
        await self.binder.bind(post_dest.user_id, pixiv_user_id)
        await self.postman.send_plain_text(message="Pixiv账号绑定成功", post_dest=post_dest)

    async def handle_bad_request(self, e: BadRequestError, post_dest: PostDestination[UID, GID]):
        pixiv_user_id = await self.binder.get_binding(post_dest.user_id)
        if pixiv_user_id is not None:
            msg = f"当前绑定账号：{pixiv_user_id}\n"
        else:
            msg = "当前未绑定Pixiv账号\n"
        msg += "命令格式：/pixivbot bind <pixiv_user_id>"
        await self.postman.send_plain_text(message=msg, post_dest=post_dest)


@context.require(CommandHandler).sub_command("unbind")
class UnbindHandler(SubCommandHandler[UID, GID], Generic[UID, GID]):
    binder = context.require(PixivAccountBinder)

    @classmethod
    def type(cls) -> str:
        return "unbind"

    @classmethod
    def enabled(cls) -> bool:
        return True

    def parse_args(self, args: Sequence[Any], post_dest: PostDestination[UID, GID]) -> dict:
        return {}

    async def actual_handle(self, *, post_dest: PostDestination[UID, GID],
                            silently: bool = False):
        await self.binder.unbind(post_dest.user_id)
        await self.postman.send_plain_text(message="Pixiv账号解绑成功", post_dest=post_dest)

    async def handle_bad_request(self, e: BadRequestError, post_dest: PostDestination[UID, GID]):
        await self.postman.send_plain_text(message="当前未绑定Pixiv账号", post_dest=post_dest)



from abc import ABC
from functools import wraps
from typing import Callable, TypeVar, Generic

from nonebot_plugin_pixivbot.postman import PostDestination

UID = TypeVar("UID")
GID = TypeVar("GID")


class Interceptor(ABC, Generic[UID, GID]):
    def __call__(self, wrapped: Callable):
        @wraps(wrapped)
        async def wrapper(post_dest: PostDestination[UID, GID],
                          silently: bool,
                          **kwargs):
            await self.intercept(wrapped, post_dest=post_dest, silently=silently, **kwargs)

        return wrapper

    async def intercept(self, wrapped_func: Callable,
                        post_dest: PostDestination[UID, GID],
                        silently: bool,
                        **kwargs):
        raise NotImplementedError()



import asyncio
import typing
from io import BytesIO
from sqlite3 import NotSupportedError

from nonebot import logger
from pixivpy_async import *
from pixivpy_async.error import TokenError

from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.data.local_tag_repo import LocalTagRepo
from nonebot_plugin_pixivbot.model import Illust, User
from nonebot_plugin_pixivbot.utils.errors import QueryError
from .abstract_repo import AbstractPixivRepo
from .compressor import Compressor
from .lazy_illust import LazyIllust
from .local_manager import CacheManager
from .pkg_context import context


def auto_retry(func):
    async def wrapped(*args, **kwargs):
        err = None
        for t in range(10):
            try:
                return await func(*args, **kwargs)
            except QueryError as e:
                raise e
            except Exception as e:
                logger.info(f"Retrying... {t + 1}/10")
                logger.exception(e)
                err = e

        raise err

    return wrapped


T = typing.TypeVar("T")


@context.register_singleton()
class RemotePixivRepo(AbstractPixivRepo):
    _conf: Config = context.require(Config)
    _local_tags = context.require(LocalTagRepo)

    def __init__(self):
        self.user_id = 0

        self.refresh_token = self._conf.pixiv_refresh_token
        self.simultaneous_query = self._conf.pixiv_simultaneous_query
        self.timeout = self._conf.pixiv_query_timeout
        self.proxy = self._conf.pixiv_proxy

        self._compressor = Compressor(enabled=self._conf.pixiv_compression_enabled,
                                      max_size=self._conf.pixiv_compression_max_size,
                                      quantity=self._conf.pixiv_compression_quantity)

        self._cache_manager = CacheManager(
            simultaneous_query=self._conf.pixiv_simultaneous_query)

    async def _refresh(self):
        # Latest app version can be found using GET /old/application-info/android
        USER_AGENT = "PixivAndroidApp/5.0.234 (Android 11; Pixel 5)"
        REDIRECT_URI = "https://app-api.pixiv.net/web/v1/users/auth/pixiv/callback"
        LOGIN_URL = "https://app-api.pixiv.net/web/v1/login"
        AUTH_TOKEN_URL = "https://oauth.secure.pixiv.net/auth/token"
        CLIENT_ID = "MOBrBDS8blbauoSck0ZfDbtuzpyT"
        CLIENT_SECRET = "lsACyCD94FhDUtGTXi3QzcFE2uU1hqtDaKeqrdwj"

        data = {
            "client_id": CLIENT_ID,
            "client_secret": CLIENT_SECRET,
            "grant_type": "refresh_token",
            "include_policy": "true",
            "refresh_token": self.refresh_token,
        }
        result = await self._papi.requests_(method="POST", url=AUTH_TOKEN_URL, data=data,
                                            headers={"User-Agent": USER_AGENT},
                                            auth=False)
        if result.has_error:
            raise TokenError(None, result)
        else:
            self._papi.set_auth(result.access_token, result.refresh_token)
            self.user_id = result["user"]["id"]

            logger.success(
                f"refresh access token successfully. new token expires in {result.expires_in} seconds.")
            logger.info(f"access_token: {result.access_token}")
            logger.info(f"refresh_token: {result.refresh_token}")

            # maybe the refresh token will be changed (even thought i haven't seen it yet)
            if result.refresh_token != self.refresh_token:
                self.refresh_token = result.refresh_token
                logger.warning(
                    f"refresh token has been changed: {result.refresh_token}")

            return result

    async def _refresh_daemon(self):
        while True:
            try:
                result = await self._refresh()
                await asyncio.sleep(result.expires_in * 0.8)
            except asyncio.CancelledError as e:
                raise e
            except Exception as e:
                logger.error(
                    "failed to refresh access token, will retry in 60s.")
                logger.exception(e)
                await asyncio.sleep(60)

    def start(self):
        self._cache_manager = CacheManager(self.simultaneous_query)
        self._pclient = PixivClient(proxy=self.proxy)
        self._papi = AppPixivAPI(client=self._pclient.start())
        self._papi.set_additional_headers({'Accept-Language': 'zh-CN'})
        self._refresh_daemon_task = asyncio.create_task(self._refresh_daemon())

    async def shutdown(self):
        await self._pclient.close()
        self._refresh_daemon_task.cancel()

    @staticmethod
    def _check_error_in_raw_result(raw_result: dict):
        if "error" in raw_result:
            raise QueryError(raw_result["error"]["user_message"]
                             or raw_result["error"]["message"] or raw_result["error"]["reason"])

    async def _flat_page(self, papi_search_func: typing.Callable[..., typing.Awaitable[dict]],
                         element_list_name: str,
                         element_mapper: typing.Optional[typing.Callable[[typing.Any], T]] = None,
                         element_filter: typing.Optional[typing.Callable[[T], bool]] = None,
                         skip: int = 0,
                         limit: int = 0,
                         limit_page: int = 0,
                         **kwargs) -> typing.List[T]:
        cur_page = 0
        items = []

        # user_bookmarks_illust 没有offset参数
        if skip:
            raw_result = await papi_search_func(offset=skip, **kwargs)
        else:
            raw_result = await papi_search_func(**kwargs)

        self._check_error_in_raw_result(raw_result)

        while (not limit or len(items) < limit) and (not limit_page or cur_page < limit_page):
            for x in raw_result[element_list_name]:
                element = x
                if element_mapper is not None:
                    element = element_mapper(x)
                if element_filter is None or element_filter(element):
                    items.append(element)
                    if len(items) >= limit:
                        break
            else:
                next_qs = AppPixivAPI.parse_qs(next_url=raw_result["next_url"])
                if next_qs is None:
                    break

                if 'viewed' in next_qs:
                    # 由于pixivpy-async的illust_recommended的bug，需要删掉这个参数
                    del next_qs['viewed']

                cur_page = cur_page + 1
                raw_result = await papi_search_func(**next_qs)
                self._check_error_in_raw_result(raw_result)

        return items

    async def _add_to_local_tags(self, illusts: typing.List[LazyIllust]):
        try:
            tags = {}
            for x in illusts:
                if isinstance(x, LazyIllust):
                    if not x.loaded:
                        continue
                    x = x.content
                for t in x.tags:
                    if t.translated_name:
                        tags[t.name] = t

            await self._local_tags.insert_many(tags.values())
        except Exception as e:
            logger.exception(e)

    async def _get_illusts(self, papi_search_func: typing.Callable[[], typing.Awaitable[dict]],
                           element_list_name: str,
                           block_tags: typing.Optional[typing.List[str]],
                           min_bookmark: int = 0,
                           min_view: int = 0,
                           skip: int = 0,
                           limit: int = 0,
                           limit_page: int = 0,
                           **kwargs):
        def illust_filter(illust: Illust) -> bool:
            # 标签过滤
            if block_tags is not None:
                for tag in block_tags:
                    if illust.has_tag(tag):
                        return False
            # 书签下限过滤
            if illust.total_bookmarks < min_bookmark:
                return False
            # 浏览量下限过滤
            if illust.total_view < min_view:
                return False
            return True

        items = await self._flat_page(papi_search_func, element_list_name,
                                      lambda x: Illust.parse_obj(x),
                                      illust_filter, skip, limit, limit_page,
                                      **kwargs)

        illusts = []
        detail_missing = 0
        for x in items:
            if "limit_unknown_360.png" in x.image_urls.large:
                detail_missing += 1
                illusts.append(LazyIllust(x.id))
            else:
                illusts.append(LazyIllust(x.id, x))

        logger.info(
            f"[remote] {len(illusts)} got, illust_detail of {detail_missing} are missed")

        if self._conf.pixiv_tag_translation_enabled:
            asyncio.create_task(self._add_to_local_tags(illusts))

        return illusts

    @auto_retry
    async def illust_detail(self, illust_id: int) -> Illust:
        logger.info(f"[remote] illust_detail {illust_id}")

        raw_result = await self._papi.illust_detail(illust_id)
        self._check_error_in_raw_result(raw_result)
        illust = Illust.parse_obj(raw_result["illust"])

        if self._conf.pixiv_tag_translation_enabled:
            asyncio.create_task(self._add_to_local_tags([illust]))

        return illust

    @auto_retry
    async def user_detail(self, user_id: int) -> User:
        logger.info(f"[remote] user_detail {user_id}")

        raw_result = await self._papi.user_detail(user_id)
        self._check_error_in_raw_result(raw_result)
        return User.parse_obj(raw_result["user"])

    @auto_retry
    async def search_illust(self, word: str, *, skip: int = 0, limit: int = 0) -> typing.List[LazyIllust]:
        if not limit:
            limit = self._conf.pixiv_random_illust_max_item
        limit_page = self._conf.pixiv_random_illust_max_page
        min_bookmark = self._conf.pixiv_random_illust_min_bookmark
        min_view = self._conf.pixiv_random_illust_min_view
        block_tags = self._conf.pixiv_block_tags

        logger.info(f"[remote] search_illust {word}")
        return await self._get_illusts(self._papi.search_illust, "illusts",
                                       block_tags, min_bookmark, min_view, skip, limit, limit_page,
                                       word=word)

    @auto_retry
    async def search_user(self, word: str, *, skip: int = 0, limit: int = 20) -> typing.List[User]:
        logger.info(f"[remote] search_user {word}")
        content = await self._flat_page(self._papi.search_user, "user_previews",
                                        lambda x: User.parse_obj(
                                            x["user"]), None,
                                        skip, limit, 1,
                                        word=word)
        return content

    @auto_retry
    async def user_illusts(self, user_id: int = 0, *, skip: int = 0, limit: int = 0) -> typing.List[LazyIllust]:
        if user_id == 0:
            user_id = self.user_id

        if not limit:
            limit = self._conf.pixiv_random_user_illust_max_item

        limit_page = self._conf.pixiv_random_user_illust_max_page
        min_bookmark = self._conf.pixiv_random_user_illust_min_bookmark
        min_view = self._conf.pixiv_random_user_illust_min_view
        block_tags = self._conf.pixiv_block_tags

        logger.info(f"[remote] user_illusts {user_id}")
        return await self._get_illusts(self._papi.user_illusts, "illusts",
                                       block_tags, min_bookmark, min_view, skip, limit, limit_page,
                                       user_id=user_id)

    @auto_retry
    async def user_bookmarks(self, user_id: int = 0, *, skip: int = 0, limit: int = 0) -> typing.List[LazyIllust]:
        if user_id == 0:
            user_id = self.user_id

        if skip:
            raise NotSupportedError("Argument skip is not supported")

        if not limit:
            limit = self._conf.pixiv_random_bookmark_max_item

        limit_page = self._conf.pixiv_random_bookmark_max_page
        min_bookmark = self._conf.pixiv_random_bookmark_min_bookmark
        min_view = self._conf.pixiv_random_bookmark_min_view
        block_tags = self._conf.pixiv_block_tags

        logger.info(f"[remote] user_bookmarks {user_id}")
        return await self._get_illusts(self._papi.user_bookmarks_illust, "illusts",
                                       block_tags, min_bookmark, min_view, skip, limit, limit_page,
                                       user_id=user_id)

    @auto_retry
    async def recommended_illusts(self, *, skip: int = 0, limit: int = 0) -> typing.List[LazyIllust]:
        if not limit:
            limit = self._conf.pixiv_random_recommended_illust_max_item

        limit_page = self._conf.pixiv_random_recommended_illust_max_page
        min_bookmark = self._conf.pixiv_random_recommended_illust_min_bookmark
        min_view = self._conf.pixiv_random_recommended_illust_min_view
        block_tags = self._conf.pixiv_block_tags

        logger.info(f"[remote] recommended_illusts")
        return await self._get_illusts(self._papi.illust_recommended, "illusts",
                                       block_tags, min_bookmark, min_view, skip, limit, limit_page)

    @auto_retry
    async def related_illusts(self, illust_id: int, *, skip: int = 0, limit: int = 0) -> typing.List[LazyIllust]:
        if not limit:
            limit = self._conf.pixiv_random_related_illust_max_item

        limit_page = self._conf.pixiv_random_related_illust_max_page
        min_bookmark = self._conf.pixiv_random_related_illust_min_bookmark
        min_view = self._conf.pixiv_random_related_illust_min_view
        block_tags = self._conf.pixiv_block_tags

        logger.info(f"[remote] related_illusts {illust_id}")
        return await self._get_illusts(self._papi.illust_related, "illusts",
                                       block_tags, min_bookmark, min_view, skip, limit, limit_page,
                                       illust_id=illust_id)

    @auto_retry
    async def illust_ranking(self, mode: str = 'day', *, skip: int = 0, limit: int = 0) -> typing.List[LazyIllust]:
        if not limit:
            limit = self._conf.pixiv_ranking_fetch_item

        block_tags = self._conf.pixiv_block_tags

        logger.info(f"[remote] illust_ranking {mode}")
        return await self._get_illusts(self._papi.illust_ranking, "illusts",
                                       block_tags, 0, 0, skip, limit, 0,
                                       mode=mode)

    @auto_retry
    async def image(self, illust: Illust) -> bytes:
        download_quantity = self._conf.pixiv_download_quantity
        custom_domain = self._conf.pixiv_download_custom_domain

        if download_quantity == "original":
            if len(illust.meta_pages) > 0:
                url = illust.meta_pages[0].image_urls.original
            else:
                url = illust.meta_single_page.original_image_url
        else:
            url = illust.image_urls.__getattribute__(download_quantity)

        if custom_domain is not None:
            url = url.replace("i.pximg.net", custom_domain)

        with BytesIO() as bio:
            await self._papi.download(url, fname=bio)
            content = bio.getvalue()
            content = await self._compressor.compress(content)
            return content



from dataclasses import dataclass
from typing import TypeVar, Generic, Optional

from nonebot_plugin_pixivbot.postman import PostDestination, PostDestinationFactory

UID = TypeVar("UID")
GID = TypeVar("GID")


@dataclass(frozen=True)
class PostIdentifier(Generic[UID, GID]):
    adapter: str
    user_id: Optional[UID]
    group_id: Optional[GID]

    def __post_init__(self):
        if not self.user_id and not self.group_id:
            raise ValueError("at least one of user_id and group_id should be not None")

    @staticmethod
    def from_post_dest(post_dest: PostDestination):
        return PostIdentifier(post_dest.adapter, post_dest.user_id, post_dest.group_id)

    def to_post_dest(self, factory: PostDestinationFactory):
        post_dest = factory.build(self.user_id, self.group_id)
        if post_dest.adapter != self.adapter:
            raise ValueError("PostDestination.adapter != PostIdentifier.adapter")
        return post_dest



from typing import Optional, Sequence, Union, TypeVar, Generic, Any

from nonebot_plugin_pixivbot.global_context import context as context
from nonebot_plugin_pixivbot.handler.common.common_handler import CommonHandler
from nonebot_plugin_pixivbot.postman import PostDestination, post_illusts
from nonebot_plugin_pixivbot.utils.decode_integer import decode_integer
from nonebot_plugin_pixivbot.utils.errors import BadRequestError

UID = TypeVar("UID")
GID = TypeVar("GID")


@context.root.register_singleton()
class RankingHandler(CommonHandler[UID, GID], Generic[UID, GID]):
    @classmethod
    def type(cls) -> str:
        return "ranking"

    mode_mapping = {"日": "day", "周": "week", "月": "month", "男性": "day_male",
                    "女性": "day_female", "原创": "week_original", "新人": "week_rookie", "漫画": "day_manga"}

    mode_reversed_mapping = {"day": "日", "week": "周", "month": "月", "day_male": "男性",
                             "day_female": "女性", "week_original": "原创", "week_rookie": "新人", "day_manga": "漫画"}

    @classmethod
    def enabled(cls) -> bool:
        return cls.conf.pixiv_ranking_query_enabled

    def validate_args(self, mode: Optional[str] = None,
                      range: Optional[Sequence[int]] = None):
        if mode and mode not in self.mode_reversed_mapping:
            raise BadRequestError(f"{mode}不是合法的榜单类型")

        if range:
            start, end = range
            if end - start + 1 > self.conf.pixiv_ranking_max_item_per_query:
                raise BadRequestError(
                    f"仅支持一次查询{self.conf.pixiv_ranking_max_item_per_query}张以下插画")
            elif start > end:
                raise BadRequestError("范围不合法")
            elif end > self.conf.pixiv_ranking_fetch_item:
                raise BadRequestError(
                    f'仅支持查询{self.conf.pixiv_ranking_fetch_item}名以内的插画')

    def parse_args(self, args: Sequence[Any], post_dest: PostDestination[UID, GID]) -> dict:
        mode = args[0] if len(args) > 0 else None
        range = args[1] if len(args) > 1 else None

        if not mode:  # 判断是不是空字符串
            mode = None
        elif mode in self.mode_mapping:
            mode = self.mode_mapping[mode]

        if not range:
            range = None
        else:
            try:
                if "-" in range:
                    start, end = range.split("-")
                    range = int(start), int(end)
                else:
                    range = decode_integer(range)
                    range = range, range
            except ValueError:
                raise BadRequestError(f"{range}不是合法的范围")

        self.validate_args(mode, range)
        return {"mode": mode, "range": range}

    async def actual_handle(self, *, mode: Optional[str] = None,
                            range: Union[Sequence[int], int, None] = None,
                            post_dest: PostDestination[UID, GID],
                            silently: bool = False):
        if mode is None:
            mode = self.conf.pixiv_ranking_default_mode

        if range is None:
            range = self.conf.pixiv_ranking_default_range

        if isinstance(range, int):
            range = range, range

        self.validate_args(mode, range)
        illusts = await self.service.illust_ranking(mode, range)
        await post_illusts(illusts,
                           header=f"这是您点的{self.mode_reversed_mapping[mode]}榜",
                           number=range[0],
                           post_dest=post_dest)



import typing

from nonebot_plugin_pixivbot.model import Illust
from .pkg_context import context


class LazyIllust:
    src: 'PixivDataSource' = None

    @classmethod
    def init_src(cls):
        # 为避免循环引用，将import推迟到get的时候
        from .repo import PixivRepo
        cls.src = context.require(PixivRepo)

    def __init__(self, id: int, content: typing.Optional[Illust] = None) -> None:
        self.id = id
        self.content = content

    async def get(self):
        if self.content is None:
            if self.src is None:
                self.init_src()
            self.content = await self.src.illust_detail(self.id)
        return self.content

    @property
    def loaded(self):
        return self.content is not None

    def __getattr__(self, attr):
        if self.content is None:
            return None
        else:
            return self.content.__getattribute__(attr)



from typing import Union, Awaitable, TypeVar, Generic, Sequence, Any

from nonebot_plugin_pixivbot.global_context import context as context
from nonebot_plugin_pixivbot.handler.cmd.command_handler import SubCommandHandler, CommandHandler
from nonebot_plugin_pixivbot.postman import PostDestination

UID = TypeVar("UID")
GID = TypeVar("GID")


@context.require(CommandHandler).sub_command("help")
class HelpHandler(SubCommandHandler[UID, GID], Generic[UID, GID]):
    help_text = """常规语句：
- 看看榜<范围>：查看pixiv榜单
- 来张图：从推荐插画随机抽选一张插画
- 来张<关键字>图：搜索关键字，从搜索结果随机抽选一张插画
- 来张<用户>老师的图：搜索画师，从该画师的插画列表里随机抽选一张插画
- 看看图<插画ID>：查看id为<插画ID>的插画
- 来张私家车：从书签中随机抽选一张插画
- 还要：重复上一次请求
- 不够色：获取上一张插画的相关推荐

命令语句：
- /pixivbot help：查看本帮助
- /pixivbot bind：绑定Pixiv账号

更多功能：参见https://github.com/ssttkkl/PixivBot
"""

    @classmethod
    def type(cls) -> str:
        return "help"

    @classmethod
    def enabled(cls) -> bool:
        return True

    def parse_args(self, args: Sequence[Any], post_dest: PostDestination[UID, GID]) \
            -> Union[dict, Awaitable[dict]]:
        return {}

    async def actual_handle(self, *, post_dest: PostDestination[UID, GID],
                            silently: bool = False):
        await self.postman.send_plain_text(self.help_text, post_dest=post_dest)



from .cmd import *
from .common import *
from .handler import Handler

__all__ = ("Handler", "CommandHandler",
           "RandomBookmarkHandler", "RandomRecommendedIllustHandler",
           "RankingHandler", "RandomIllustHandler",
           "RandomUserIllustHandler", "IllustHandler",
           "MoreHandler", "RandomRelatedIllustHandler")



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from nonebot_plugin_pixivbot.postman.model import IllustMessageModel, IllustMessagesModel
from nonebot_plugin_pixivbot.postman.post_destination import PostDestination

UID = TypeVar("UID")
GID = TypeVar("GID")


class Postman(ABC, Generic[UID, GID]):
    @abstractmethod
    async def send_plain_text(self, message: str,
                              *, post_dest: PostDestination[UID, GID]):
        raise NotImplementedError()

    @abstractmethod
    async def send_illust(self, model: IllustMessageModel,
                          *, post_dest: PostDestination[UID, GID]):
        raise NotImplementedError()

    @abstractmethod
    async def send_illusts(self, model: IllustMessagesModel,
                           *, post_dest: PostDestination[UID, GID]):
        raise NotImplementedError()


__all__ = ("Postman",)



import time
from collections import OrderedDict
from typing import Optional, TypeVar, Generic

from nonebot import logger

from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.global_context import context as context
from nonebot_plugin_pixivbot.model import PostIdentifier
from nonebot_plugin_pixivbot.postman import PostDestination

UID = TypeVar("UID")
GID = TypeVar("GID")

PD = PostDestination[UID, GID]
ID = PostIdentifier[UID, GID]


class Req(Generic[UID, GID]):
    def __init__(self, handler: 'Handler[UID, GID]',
                 *args, **kwargs):
        self.timestamp = 0
        self.handler = handler
        self.args = args
        self.kwargs = kwargs
        self.refresh()

    def refresh(self):
        self.timestamp = time.time()

    def __call__(self, *, post_dest: PD, **kwargs):
        actual_kwargs = self.kwargs
        for x, y in kwargs.items():
            actual_kwargs[x] = y

        return self.handler.handle(*self.args, post_dest=post_dest, **actual_kwargs)


class Resp:
    def __init__(self, illust_id: int):
        self.timestamp = 0
        self.illust_id = illust_id
        self.refresh()

    def refresh(self):
        self.timestamp = time.time()


@context.register_singleton()
class Recorder(Generic[UID, GID]):
    conf = context.require(Config)

    def __init__(self):
        self._reqs = OrderedDict[ID, Req[UID, GID]]()
        self._resps = OrderedDict[ID, Resp]()

    @staticmethod
    def _key_fallback(key: ID) -> Optional[ID]:
        if key.user_id is not None and key.group_id is not None:
            return ID(None, key.group_id)
        else:
            return None

    def _pop_expired_req(self):
        now = time.time()
        while len(self._reqs) > 0:
            key, req = next(iter(self._reqs.items()))
            if now - req.timestamp > self.conf.pixiv_query_expires_in:
                self._reqs.popitem(last=False)
                logger.info(f"expired req popped: ({key})")
            else:
                break

    def record_req(self, record: Req, key: ID):
        self._pop_expired_req()
        if key in self._reqs:
            self._reqs.move_to_end(key)
        self._reqs[key] = record

    def get_req(self, key: ID) -> Optional[Req]:
        self._pop_expired_req()
        if key in self._reqs:
            req = self._reqs[key]
            req.refresh()
            self._reqs.move_to_end(key)
            return req
        else:
            key = self._key_fallback(key)
            if key is not None:
                return self.get_req(key)
            else:
                return None

    def _pop_expired_resp(self):
        now = time.time()
        while len(self._resps) > 0:
            key, rec = next(iter(self._resps.items()))
            if now - rec.timestamp > self.conf.pixiv_query_expires_in:
                self._resps.popitem(last=False)
                logger.info(f"expired illust popped: ({key})")
            else:
                break

    def record_resp(self, illust_id: int, key: ID):
        self._pop_expired_resp()
        if key in self._resps:
            self._resps.move_to_end(key)
        self._resps[key] = Resp(illust_id)

    def get_resp(self, key: ID) -> Optional[int]:
        self._pop_expired_resp()
        if key in self._resps:
            rec = self._resps[key]
            return rec.illust_id
        else:
            key = self._key_fallback(key)
            if key is not None:
                return self.get_resp(key)
            else:
                return None



from typing import TypeVar, Generic, AsyncGenerator

from pymongo import ReturnDocument

from nonebot_plugin_pixivbot.data.source import MongoDataSource
from nonebot_plugin_pixivbot.global_context import context as context
from nonebot_plugin_pixivbot.model import Subscription
from nonebot_plugin_pixivbot.model.post_identifier import PostIdentifier

UID = TypeVar("UID")
GID = TypeVar("GID")

ID = PostIdentifier[UID, GID]


@context.register_singleton()
class SubscriptionRepo(Generic[UID, GID]):
    mongo = context.require(MongoDataSource)

    async def get(self, identifier: ID) -> AsyncGenerator[Subscription, None]:
        if identifier.group_id:
            query = {"adapter": identifier.adapter, "group_id": identifier.group_id}
        elif identifier.user_id:
            query = {"adapter": identifier.adapter, "user_id": identifier.user_id}
        else:
            raise ValueError("at least one of user_id and group_id should be not None")

        async for obj in self.mongo.db.subscription.find(query):
            yield Subscription.parse_obj(obj)

    async def get_all(self, adapter: str) -> AsyncGenerator[Subscription, None]:
        query = {"adapter": adapter}

        async for obj in self.mongo.db.subscription.find(query):
            yield Subscription.parse_obj(obj)

    async def update(self, subscription: Subscription) -> Subscription:
        if subscription.group_id:
            query = {"adapter": subscription.adapter, "group_id": subscription.group_id, "type": subscription.type}
        elif subscription.user_id:
            query = {"adapter": subscription.adapter, "user_id": subscription.user_id, "type": subscription.type}
        else:
            raise ValueError("at least one of user_id and group_id should be not None")

        return await self.mongo.db.subscription.find_one_and_replace(query, subscription.dict(),
                                                                     return_document=ReturnDocument.BEFORE,
                                                                     upsert=True)

    async def delete(self, identifier: ID, type: str):
        if identifier.group_id:
            query = {"adapter": identifier.adapter, "group_id": identifier.group_id, "type": type}
        elif identifier.user_id:
            query = {"adapter": identifier.adapter, "user_id": identifier.user_id, "type": type}
        else:
            raise ValueError("at least one of user_id and group_id should be not None")

        if type == 'all':
            del query["type"]
            await self.mongo.db.subscription.delete_many(query)
        else:
            await self.mongo.db.subscription.delete_one(query)


__all__ = ("SubscriptionRepo",)



from typing import TypeVar, Generic, Sequence, Any

from nonebot import get_bot

from nonebot_plugin_pixivbot.data.pixiv import PixivRepo
from nonebot_plugin_pixivbot.global_context import context as context
from nonebot_plugin_pixivbot.handler.cmd.command_handler import SubCommandHandler, CommandHandler
from nonebot_plugin_pixivbot.postman import PostDestination

UID = TypeVar("UID")
GID = TypeVar("GID")


@context.require(CommandHandler).sub_command("invalidate_cache")
class InvalidateCacheHandler(SubCommandHandler[UID, GID], Generic[UID, GID]):
    pixiv_data_source = context.require(PixivRepo)

    @classmethod
    def type(cls) -> str:
        return "invalidate_cache"

    @classmethod
    def enabled(cls) -> bool:
        return True

    def parse_args(self, args: Sequence[Any], post_dest: PostDestination[UID, GID]) -> dict:
        return {}

    async def actual_handle(self, *, post_dest: PostDestination[UID, GID],
                            silently: bool = False):
        bot = get_bot()
        if post_dest.user_id not in bot.config.superusers:
            await self.postman.send_plain_text(message="只有超级用户可以调用该命令", post_dest=post_dest)
            return

        await self.pixiv_data_source.invalidate_cache()
        await self.postman.send_plain_text(message="ok", post_dest=post_dest)



from typing import TypeVar, Generic, Sequence, Dict, Any, Optional

from pydantic import validator
from pydantic.generics import GenericModel

from nonebot_plugin_pixivbot.model import PostIdentifier

UID = TypeVar("UID")
GID = TypeVar("GID")


class Subscription(GenericModel, Generic[UID, GID]):
    adapter: str
    user_id: Optional[UID]
    group_id: Optional[GID]
    type: str
    kwargs: Dict[str, Any]
    schedule: Sequence[int]

    @validator("user_id")
    def validate(cls, user_id, values, **kwargs):
        group_id = None
        if "group_id" in values:
            group_id = values["group_id"]

        if not user_id and not group_id:
            raise ValueError("at least one of user_id and group_id should be not None")
        return user_id

    @property
    def identifier(self):
        return PostIdentifier(self.adapter, self.user_id, self.group_id)


__all__ = ("Subscription",)



import re
from datetime import datetime, timedelta
from inspect import isawaitable
from typing import TypeVar, Dict, List, Sequence, Union, Optional, Generic

from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.triggers.interval import IntervalTrigger
from lazy import lazy
from nonebot import logger, get_driver

from nonebot_plugin_pixivbot.data.subscription_repo import SubscriptionRepo
from nonebot_plugin_pixivbot.global_context import context as context
from nonebot_plugin_pixivbot.model import Subscription, PostIdentifier
from nonebot_plugin_pixivbot.postman import PostDestinationFactory
from nonebot_plugin_pixivbot.utils.errors import BadRequestError
from nonebot_plugin_pixivbot.utils.nonebot import get_adapter_name

UID = TypeVar("UID")
GID = TypeVar("GID")

ID = PostIdentifier[UID, GID]


@context.register_singleton()
class Scheduler(Generic[UID, GID]):
    apscheduler = context.require(AsyncIOScheduler)
    subscriptions = context.require(SubscriptionRepo)
    post_dest_factory = context.require(PostDestinationFactory)

    @staticmethod
    def _make_job_id(type: str, identifier: ID):
        if identifier.group_id:
            return f'scheduled {type} g{identifier.group_id}'
        else:
            return f'scheduled {type} u{identifier.user_id}'

    @staticmethod
    def _parse_schedule(raw_schedule: str) -> Sequence[int]:
        start_only_mat = re.fullmatch(r'(\d+):(\d+)', raw_schedule)
        if start_only_mat is not None:
            g = start_only_mat.groups()
            start_hour, start_minute = int(g[0]), int(g[1])
            interval_hour, interval_minute = 24, 0
        else:
            interval_only_mat = re.fullmatch(
                r'(\d+):(\d+)\*x', raw_schedule)
            if interval_only_mat is not None:
                g = interval_only_mat.groups()
                start_hour, start_minute = 0, 0
                interval_hour, interval_minute = int(g[0]), int(g[1])
            else:
                mat = re.fullmatch(r'(\d+):(\d+)\+(\d+):(\d+)\*x', raw_schedule)
                if mat is not None:
                    g = mat.groups()
                    start_hour, start_minute = int(g[0]), int(g[1])
                    interval_hour, interval_minute = int(g[2]), int(g[3])
                else:
                    raise BadRequestError(f'{raw_schedule}不是合法的时间')

        if start_hour < 0 or start_hour >= 24 or start_minute < 0 or start_minute >= 60 \
                or interval_hour < 0 or interval_minute >= 60 \
                or (interval_hour > 0 and interval_minute < 0) or (interval_hour == 0 and interval_minute <= 0):
            raise BadRequestError(f'{raw_schedule}不是合法的时间')

        return start_hour, start_minute, interval_hour, interval_minute

    @lazy
    def _handlers(self) -> Dict[str, 'Handler']:
        # 解决Handler和Scheduler的循环引用
        from nonebot_plugin_pixivbot.handler import RandomBookmarkHandler, RandomRecommendedIllustHandler, \
            RankingHandler, RandomIllustHandler, RandomUserIllustHandler
        return {
            RandomBookmarkHandler.type(): context.require(RandomBookmarkHandler),
            RandomRecommendedIllustHandler.type(): context.require(RandomRecommendedIllustHandler),
            RankingHandler.type(): context.require(RankingHandler),
            RandomIllustHandler.type(): context.require(RandomIllustHandler),
            RandomUserIllustHandler.type(): context.require(RandomUserIllustHandler),
        }

    def _add_job(self, sub: Subscription[UID, GID]):
        offset_hour, offset_minute, hours, minutes = sub.schedule
        trigger = IntervalTrigger(hours=hours, minutes=minutes,
                                  start_date=datetime.now().replace(hour=offset_hour, minute=offset_minute,
                                                                    second=0, microsecond=0) + timedelta(days=-1))

        identifier = sub.identifier
        job_id = self._make_job_id(sub.type, identifier)
        post_dest = identifier.to_post_dest(self.post_dest_factory)
        self.apscheduler.add_job(self._handlers[sub.type].handle, id=job_id, trigger=trigger,
                                 kwargs={"post_dest": post_dest, "silently": True, **sub.kwargs})
        logger.success(f"scheduled {job_id} {trigger}")

    def _remove_job(self, type: str, identifier: ID):
        job_id = self._make_job_id(type, identifier)
        self.apscheduler.remove_job(job_id)
        logger.success(f"unscheduled {job_id}")

    async def start(self):
        async for subscription in self.subscriptions.get_all(get_adapter_name()):
            self._add_job(subscription)

    async def stop(self):
        jobs = self.apscheduler.get_jobs()
        for j in jobs:
            if j.id.startswith("scheduled"):
                j.remove()

    async def schedule(self, type: str,
                       schedule: Union[str, Sequence[int]],
                       args: Optional[list] = None,
                       *, identifier: PostIdentifier[UID, GID]):
        if type not in self._handlers:
            raise BadRequestError(f"{type}不是合法的类型")

        if args is None:
            args = []

        if isinstance(schedule, str):
            schedule = self._parse_schedule(schedule)

        kwargs = self._handlers[type].parse_args(args, identifier.to_post_dest(self.post_dest_factory))
        if isawaitable(kwargs):
            kwargs = await kwargs

        sub = Subscription(adapter=identifier.adapter,
                           user_id=identifier.user_id,
                           group_id=identifier.group_id,
                           type=type,
                           schedule=schedule,
                           kwargs=kwargs)

        old_sub = await self.subscriptions.update(sub)
        if old_sub is not None:
            self._remove_job(sub.type, sub.identifier)
        self._add_job(sub)

    async def unschedule(self, type: str,
                         identifier: PostIdentifier[UID, GID]):
        if type == "all":
            async for sub in self.subscriptions.get(identifier):
                self._remove_job(sub.type, identifier)
        elif type in self._handlers:
            self._remove_job(type, identifier)
        else:
            raise BadRequestError(f"{type}不是合法的类型")

        await self.subscriptions.delete(identifier, type)

    async def all_subscription(self, identifier: PostIdentifier[UID, GID]) -> List[dict]:
        return [x async for x in self.subscriptions.get(identifier)]


scheduler = context.require(Scheduler)
get_driver().on_bot_connect(scheduler.start)
get_driver().on_bot_disconnect(scheduler.stop)

__all__ = ("Scheduler",)



from .bind_handler import BindHandler, UnbindHandler
from .command_handler import CommandHandler
from .help_handler import HelpHandler
from .invalidate_cache_handler import InvalidateCacheHandler
from .schedule_handler import ScheduleHandler, UnscheduleHandler

__all__ = ("CommandHandler",)



from abc import ABC
from abc import abstractmethod
from typing import Type, Callable, Union, Awaitable, TypeVar, Generic, Sequence, Any

from nonebot_plugin_pixivbot.global_context import context as context
from nonebot_plugin_pixivbot.handler.handler import Handler
from nonebot_plugin_pixivbot.handler.interceptor.default_error_interceptor import DefaultErrorInterceptor
from nonebot_plugin_pixivbot.postman import PostDestination
from nonebot_plugin_pixivbot.utils.errors import BadRequestError

UID = TypeVar("UID")
GID = TypeVar("GID")


class SubCommandHandler(Handler[UID, GID], ABC, Generic[UID, GID]):
    @abstractmethod
    def parse_args(self, args: Sequence[Any], post_dest: PostDestination[UID, GID]) -> Union[dict, Awaitable[dict]]:
        raise NotImplementedError()

    async def handle_bad_request(self, e: BadRequestError, post_dest: PostDestination[UID, GID]):
        self.postman.send_plain_text(e.message, post_dest=post_dest)


@context.register_singleton()
class CommandHandler(Handler[UID, GID], Generic[UID, GID]):
    interceptor = context.require(DefaultErrorInterceptor)

    def __init__(self):
        self.handlers = dict[str, Type[SubCommandHandler[UID, GID]]]()

    @classmethod
    def type(cls) -> str:
        return "command"

    @classmethod
    def enabled(cls) -> bool:
        return True

    def sub_command(self, type: str) \
            -> Callable[[Type[SubCommandHandler[UID, GID]]], Type[SubCommandHandler[UID, GID]]]:
        def decorator(cls: Type[SubCommandHandler[UID, GID]]):
            context.register_singleton()(cls)
            self.handlers[type] = cls
            return cls

        return decorator

    def parse_args(self, args: Sequence[Any], post_dest: PostDestination[UID, GID]) -> dict:
        return {"args": args[0]}

    async def actual_handle(self, *, args: Sequence[Any],
                            post_dest: PostDestination[UID, GID],
                            silently: bool = False):
        if len(args) == 0:
            handler = context.require(self.handlers["help"])
        else:
            handler = context.require(self.handlers[args[0]])

        try:
            await handler.handle(*args[1:], post_dest=post_dest)
        except BadRequestError as e:
            await handler.handle_bad_request(e, post_dest)



import asyncio
import typing


class CacheManager:
    def __init__(self, simultaneous_query: int = 4):
        self._semaphore = asyncio.Semaphore(
            value=simultaneous_query)  # 用于限制从远程获取的并发量
        self._waiting = {}

    T = typing.TypeVar("T")

    async def get(self, identifier: typing.Any,
                  cache_loader: typing.Callable[[], typing.Coroutine[typing.Any, typing.Any, T]],
                  remote_fetcher: typing.Callable[[], typing.Coroutine[typing.Any, typing.Any, T]],
                  cache_updater: typing.Callable[[T], typing.Coroutine[typing.Any, typing.Any, typing.NoReturn]],
                  hook_on_cache: typing.Optional[typing.Callable[[T], T]] = None,
                  hook_on_fetch: typing.Optional[typing.Callable[[T], T]] = None,
                  timeout: typing.Optional[int] = 0) -> T:
        cache = await cache_loader()
        if cache is not None:
            if hook_on_cache:
                cache = hook_on_cache(cache)
            return cache

        if identifier in self._waiting:
            return await self._waiting[identifier]

        fut = asyncio.Future()
        self._waiting[identifier] = fut
        try:
            asyncio.create_task(self._fetch(
                fut, remote_fetcher, cache_updater, timeout))
            result = await fut
            if hook_on_fetch:
                result = hook_on_fetch(result)
            return result
        finally:
            await self._waiting.pop(identifier)

    async def _fetch(self, fut: asyncio.Future,
                     remote_fetcher: typing.Callable[[], typing.Coroutine[typing.Any, typing.Any, T]],
                     cache_updater: typing.Callable[[T], typing.Coroutine[typing.Any, typing.Any, typing.NoReturn]],
                     timeout: typing.Optional[float] = None):
        await self._semaphore.acquire()
        try:
            result = await asyncio.wait_for(remote_fetcher(), timeout)
            await cache_updater(result)
            fut.set_result(result)
        except Exception as e:
            fut.set_exception(e)
        finally:
            self._semaphore.release()


__all__ = ("CacheManager",)



from typing import TypeVar, Generic, Sequence, Any

from nonebot_plugin_pixivbot.global_context import context as context
from nonebot_plugin_pixivbot.handler.common.common_handler import CommonHandler
from nonebot_plugin_pixivbot.postman import post_illust
from nonebot_plugin_pixivbot.postman.post_destination import PostDestination
from nonebot_plugin_pixivbot.utils.errors import BadRequestError

UID = TypeVar("UID")
GID = TypeVar("GID")


@context.root.register_singleton()
class IllustHandler(CommonHandler[UID, GID], Generic[UID, GID]):
    @classmethod
    def type(cls) -> str:
        return "illust"

    @classmethod
    def enabled(cls) -> bool:
        return cls.conf.pixiv_illust_query_enabled

    def parse_args(self, args: Sequence[Any], post_dest: PostDestination[UID, GID]) -> dict:
        try:
            return {"illust_id": int(args[0])}
        except ValueError:
            raise BadRequestError(f"{args[0]}不是合法的插画ID")

    async def actual_handle(self, *, illust_id: int,
                            post_dest: PostDestination[UID, GID],
                            silently: bool = False):
        illust = await self.service.illust_detail(illust_id)
        await post_illust(illust, post_dest=post_dest)



import typing
from functools import partial

from nonebot import get_driver

from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.model import Illust, User
from .abstract_repo import AbstractPixivRepo
from .lazy_illust import LazyIllust
from .local_manager import CacheManager
from .local_repo import LocalPixivRepo
from .pkg_context import context
from .remote_repo import RemotePixivRepo


def do_skip_and_limit(items: list, skip: int, limit: int) -> list:
    if skip:
        if limit and len(items) > skip + limit:
            return items[skip:skip + limit]
        else:
            return items[skip:]
    elif limit and len(items) > limit:
        return items[:limit]
    else:
        return items


@context.root.register_singleton()
class PixivRepo(AbstractPixivRepo):
    remote: RemotePixivRepo = context.require(RemotePixivRepo)
    cache: LocalPixivRepo = context.require(LocalPixivRepo)

    _conf: Config = context.require(Config)
    timeout = _conf.pixiv_query_timeout

    def __init__(self):
        self._cache_manager = None

    def start(self):
        self.remote.start()
        self._cache_manager = CacheManager()

    async def shutdown(self):
        await self.remote.shutdown()

    def invalidate_cache(self):
        return self.cache.invalidate_cache()

    async def illust_detail(self, illust_id: int) -> Illust:
        return await self._cache_manager.get(
            identifier=(6, illust_id),
            cache_loader=partial(self.cache.illust_detail,
                                 illust_id=illust_id),
            remote_fetcher=partial(
                self.remote.illust_detail, illust_id=illust_id),
            cache_updater=self.cache.update_illust_detail,
            timeout=self.timeout
        )

    async def user_detail(self, user_id: int) -> User:
        return await self._cache_manager.get(
            identifier=(9, user_id),
            cache_loader=partial(self.cache.user_detail,
                                 user_id=user_id),
            remote_fetcher=partial(
                self.remote.user_detail, user_id=user_id),
            cache_updater=self.cache.update_user_detail,
            timeout=self.timeout
        )

    async def search_illust(self, word: str, *, skip: int = 0, limit: int = 0) -> typing.List[LazyIllust]:
        # skip和limit只作用于cache_loader
        return await self._cache_manager.get(
            identifier=(0, word),
            cache_loader=partial(self.cache.search_illust,
                                 word=word, skip=skip, limit=limit),
            remote_fetcher=partial(self.remote.search_illust, word=word),
            cache_updater=lambda content: self.cache.update_search_illust(
                word, content),
            hook_on_fetch=lambda result: do_skip_and_limit(
                result, skip, limit),
            timeout=self.timeout
        )

    async def search_user(self, word: str, *, skip: int = 0, limit: int = 0) -> typing.List[User]:
        return await self._cache_manager.get(
            identifier=(1, word),
            cache_loader=partial(self.cache.search_user,
                                 word=word, skip=skip, limit=limit),
            remote_fetcher=partial(self.remote.search_user, word=word),
            cache_updater=lambda content: self.cache.update_search_user(
                word, content),
            hook_on_fetch=lambda result: do_skip_and_limit(
                result, skip, limit),
            timeout=self.timeout
        )

    async def user_illusts(self, user_id: int = 0, *, skip: int = 0, limit: int = 0) -> typing.List[LazyIllust]:
        return await self._cache_manager.get(
            identifier=(2, user_id),
            cache_loader=partial(self.cache.user_illusts,
                                 user_id=user_id, skip=skip, limit=limit),
            remote_fetcher=partial(self.remote.user_illusts, user_id=user_id),
            cache_updater=lambda content: self.cache.update_user_illusts(
                user_id, content),
            hook_on_fetch=lambda result: do_skip_and_limit(
                result, skip, limit),
            timeout=self.timeout
        )

    async def user_bookmarks(self, user_id: int = 0, *, skip: int = 0, limit: int = 0) -> typing.List[LazyIllust]:
        return await self._cache_manager.get(
            identifier=(3, user_id),
            cache_loader=partial(self.cache.user_bookmarks,
                                 user_id=user_id, skip=skip, limit=limit),
            remote_fetcher=partial(
                self.remote.user_bookmarks, user_id=user_id),
            cache_updater=lambda content: self.cache.update_user_bookmarks(
                user_id, content),
            hook_on_fetch=lambda result: do_skip_and_limit(
                result, skip, limit),
            timeout=self.timeout
        )

    async def recommended_illusts(self, *, skip: int = 0, limit: int = 0) -> typing.List[LazyIllust]:
        return await self._cache_manager.get(
            identifier=(4,),
            cache_loader=partial(
                self.cache.recommended_illusts, skip=skip, limit=limit),
            remote_fetcher=self.remote.recommended_illusts,
            cache_updater=self.cache.update_recommended_illusts,
            hook_on_fetch=lambda result: do_skip_and_limit(
                result, skip, limit),
            timeout=self.timeout
        )

    async def related_illusts(self, illust_id: int, *, skip: int = 0, limit: int = 0) -> typing.List[LazyIllust]:
        return await self._cache_manager.get(
            identifier=(8, illust_id),
            cache_loader=partial(
                self.cache.related_illusts, illust_id=illust_id, skip=skip, limit=limit),
            remote_fetcher=partial(
                self.remote.related_illusts, illust_id=illust_id),
            cache_updater=lambda content: self.cache.update_related_illusts(
                illust_id, content),
            hook_on_fetch=lambda result: do_skip_and_limit(
                result, skip, limit),
            timeout=self.timeout
        )

    async def illust_ranking(self, mode: str = 'day', *, skip: int = 0, limit: int = 0) -> typing.List[LazyIllust]:
        return await self._cache_manager.get(
            identifier=(5, mode),
            cache_loader=partial(
                self.cache.illust_ranking, mode=mode, skip=skip, limit=limit),
            remote_fetcher=partial(
                self.remote.illust_ranking, mode=mode),
            cache_updater=lambda content: self.cache.update_illust_ranking(
                mode, content),
            hook_on_fetch=lambda result: do_skip_and_limit(
                result, skip, limit),
            timeout=self.timeout
        )

    async def image(self, illust: Illust) -> bytes:
        return await self._cache_manager.get(
            identifier=(7, illust.id),
            cache_loader=partial(self.cache.image, illust=illust),
            remote_fetcher=partial(self.remote.image, illust=illust),
            cache_updater=lambda content: self.cache.update_image(
                illust, content),
            timeout=self.timeout
        )


pixiv_data_source = context.require(PixivRepo)

get_driver().on_startup(pixiv_data_source.start)
get_driver().on_shutdown(pixiv_data_source.shutdown)

__all__ = ('PixivRepo',)



from typing import List, TypeVar, Generic, Sequence, Any

from nonebot_plugin_pixivbot.global_context import context as context
from nonebot_plugin_pixivbot.handler.cmd.command_handler import CommandHandler, SubCommandHandler
from nonebot_plugin_pixivbot.model import PostIdentifier
from nonebot_plugin_pixivbot.postman import PostDestination
from nonebot_plugin_pixivbot.service.scheduler import Scheduler
from nonebot_plugin_pixivbot.utils.errors import BadRequestError

UID = TypeVar("UID")
GID = TypeVar("GID")


@context.require(CommandHandler).sub_command("schedule")
class ScheduleHandler(SubCommandHandler[UID, GID], Generic[UID, GID]):
    scheduler = context.require(Scheduler)

    @classmethod
    def type(cls) -> str:
        return "schedule"

    @classmethod
    def enabled(cls) -> bool:
        return True

    def parse_args(self, args: Sequence[Any], post_dest: PostIdentifier[UID, GID]) -> dict:
        return {"type": args[0],
                "schedule": args[1],
                "args": args[2:]}

    async def actual_handle(self, *, type: str,
                            schedule: str,
                            args: List,
                            post_dest: PostDestination[UID, GID],
                            silently: bool = False):
        await self.scheduler.schedule(type, schedule, args, identifier=PostIdentifier.from_post_dest(post_dest))
        await self.postman.send_plain_text(message="订阅成功", post_dest=post_dest)

    async def handle_bad_request(self, e: BadRequestError, post_dest: PostDestination[UID, GID]):
        subscription = await self.scheduler.all_subscription(PostIdentifier.from_post_dest(post_dest))
        msg = "当前订阅：\n"
        if len(subscription) > 0:
            for x in subscription:
                msg += f'{x.type} {str(x.schedule[0]).zfill(2)}:{str(x.schedule[1]).zfill(2)}+{str(x.schedule[2]).zfill(2)}:{str(x.schedule[3]).zfill(2)}*x\n'
        else:
            msg += '无\n'
        msg += "\n"
        msg += "命令格式：/pixivbot schedule <type> <schedule> <..args>\n"
        msg += "示例：/pixivbot schedule ranking 06:00*x day 1-5\n"
        await self.postman.send_plain_text(message=msg, post_dest=post_dest)


@context.require(CommandHandler).sub_command("unschedule")
class UnscheduleHandler(SubCommandHandler[UID, GID], Generic[UID, GID]):
    scheduler = context.require(Scheduler)

    @classmethod
    def type(cls) -> str:
        return "unschedule"

    @classmethod
    def enabled(cls) -> bool:
        return True

    def parse_args(self, args: Sequence[Any], post_dest: PostIdentifier[UID, GID]) -> dict:
        return {"type": args[0]}

    async def actual_handle(self, *, type: str,
                            post_dest: PostDestination[UID, GID],
                            silently: bool = False):
        await self.scheduler.unschedule(type, PostIdentifier.from_post_dest(post_dest))
        await self.postman.send_plain_text(message="取消订阅成功", post_dest=post_dest)

    async def handle_bad_request(self, e: BadRequestError, post_dest: PostDestination[UID, GID]):
        subscription = await self.scheduler.all_subscription(PostIdentifier.from_post_dest(post_dest))
        msg = "当前订阅：\n"
        if len(subscription) > 0:
            for x in subscription:
                msg += f'{x.type} {str(x.schedule[0]).zfill(2)}:{str(x.schedule[1]).zfill(2)}+{str(x.schedule[2]).zfill(2)}:{str(x.schedule[3]).zfill(2)}*x\n'
        else:
            msg += '无\n'
        msg += "\n"
        msg += "命令格式：/pixivbot unschedule <type>"
        await self.postman.send_plain_text(message=msg, post_dest=post_dest)



from typing import TypeVar, Generic

from nonebot_plugin_pixivbot.global_context import context as context
from nonebot_plugin_pixivbot.handler.common.common_handler import CommonHandler
from nonebot_plugin_pixivbot.model import PostIdentifier
from nonebot_plugin_pixivbot.postman import PostDestination
from nonebot_plugin_pixivbot.utils.errors import BadRequestError

UID = TypeVar("UID")
GID = TypeVar("GID")


@context.root.register_singleton()
class MoreHandler(CommonHandler[UID, GID], Generic[UID, GID]):
    @classmethod
    def type(cls) -> str:
        return "more"

    @classmethod
    def enabled(cls) -> bool:
        return cls.conf.pixiv_more_enabled

    async def actual_handle(self, *, count: int = 1,
                            post_dest: PostDestination[UID, GID],
                            silently: bool = False):
        req = self.recorder.get_req(PostIdentifier.from_post_dest(post_dest))
        if not req:
            raise BadRequestError("你还没有发送过请求")

        await req(count=count, post_dest=post_dest)



"""
nonebot-plugin-pixivbot

@Author         : ssttkkl
@License        : MIT
@GitHub         : https://github.com/ssttkkl/nonebot-plugin-pixivbot
"""

from .global_context import context
from .providers import *

__all__ = ("context",)



from abc import ABC, abstractmethod
from inspect import isawaitable
from typing import Awaitable, Union, TypeVar, Generic, Optional, Sequence, Any

from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.global_context import context as context
from nonebot_plugin_pixivbot.handler.interceptor.interceptor import Interceptor
from nonebot_plugin_pixivbot.postman import Postman
from nonebot_plugin_pixivbot.postman.post_destination import PostDestination
from nonebot_plugin_pixivbot.utils.errors import BadRequestError

UID = TypeVar("UID")
GID = TypeVar("GID")


class Handler(ABC, Generic[UID, GID]):
    conf = context.require(Config)
    postman = context.require(Postman)

    interceptor: Optional[Interceptor] = None

    @classmethod
    @abstractmethod
    def type(cls) -> str:
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def enabled(cls) -> bool:
        raise NotImplementedError()

    def parse_args(self, args: Sequence[Any], post_dest: PostDestination[UID, GID]) -> Union[dict, Awaitable[dict]]:
        """
        将位置参数转化为命名参数
        :param args: 位置参数sequence
        :param post_dest: PostDestination
        :return: 命名参数dict
        """
        return {}

    async def handle(self, *args,
                     post_dest: PostDestination[UID, GID],
                     silently: bool = False,
                     **kwargs):
        try:
            parsed_kwargs = self.parse_args(args, post_dest)
            if isawaitable(parsed_kwargs):
                parsed_kwargs = await parsed_kwargs
        except:
            raise BadRequestError("参数错误")

        kwargs = {**kwargs, **parsed_kwargs}

        if self.interceptor is not None:
            await self.interceptor.intercept(self.actual_handle,
                                             post_dest=post_dest, silently=silently,
                                             **kwargs)
        else:
            await self.actual_handle(post_dest=post_dest, silently=silently, **kwargs)

    @abstractmethod
    async def actual_handle(self, post_dest: PostDestination[UID, GID],
                            silently: bool = False, **kwargs):
        """
        处理指令
        :param post_dest: PostDestination
        :param silently: 失败时不发送消息
        :param kwargs: 参数dict
        """
        raise NotImplementedError()



import typing
from abc import ABC, abstractmethod

from nonebot_plugin_pixivbot.model import Illust, User
from .lazy_illust import LazyIllust


class AbstractPixivRepo(ABC):
    @abstractmethod
    async def illust_detail(self, illust_id: int) -> Illust:
        raise NotImplementedError()

    @abstractmethod
    async def user_detail(self, user_id: int) -> User:
        raise NotImplementedError()

    @abstractmethod
    async def search_illust(self, word: str, *, skip: int = 0, limit: int = 0) -> typing.List[LazyIllust]:
        raise NotImplementedError()

    @abstractmethod
    async def search_user(self, word: str, *, skip: int = 0, limit: int = 0) -> typing.List[User]:
        raise NotImplementedError()

    @abstractmethod
    async def user_illusts(self, user_id: int = 0, *, skip: int = 0, limit: int = 0) -> typing.List[LazyIllust]:
        raise NotImplementedError()

    @abstractmethod
    async def user_bookmarks(self, user_id: int = 0, *, skip: int = 0, limit: int = 0) -> typing.List[LazyIllust]:
        raise NotImplementedError()

    @abstractmethod
    async def recommended_illusts(self, *, skip: int = 0, limit: int = 0) -> typing.List[LazyIllust]:
        raise NotImplementedError()

    @abstractmethod
    async def related_illusts(self, illust_id: int, *, skip: int = 0, limit: int = 0) -> typing.List[LazyIllust]:
        raise NotImplementedError()

    @abstractmethod
    async def illust_ranking(self, mode: str = 'day', *, skip: int = 0, limit: int = 0) -> typing.List[LazyIllust]:
        raise NotImplementedError()

    @abstractmethod
    async def image(self, illust: Illust) -> bytes:
        raise NotImplementedError()



from typing import TypeVar, Generic, Optional

from nonebot_plugin_pixivbot.data import PixivBindingRepo
from nonebot_plugin_pixivbot.global_context import context as context
from nonebot_plugin_pixivbot.model.pixiv_binding import PixivBinding
from nonebot_plugin_pixivbot.utils.nonebot import get_adapter_name

UID = TypeVar("UID")


@context.register_singleton()
class PixivAccountBinder(Generic[UID]):
    repo = context.require(PixivBindingRepo)

    async def bind(self, user_id: UID, pixiv_user_id: int):
        binding = PixivBinding(adapter=get_adapter_name(), user_id=user_id, pixiv_user_id=pixiv_user_id)
        await self.repo.update(binding)

    async def unbind(self, user_id: UID) -> bool:
        return await self.repo.remove(get_adapter_name(), user_id)

    async def get_binding(self, user_id: UID) -> Optional[int]:
        binding = await self.repo.get(get_adapter_name(), user_id)
        return binding["pixiv_user_id"]



from .illust import Illust
from .post_identifier import PostIdentifier
from .subscription import Subscription
from .tag import Tag
from .user import User

__all__ = ("Illust", "User", "Tag", "Subscription", "PostIdentifier")



import typing
from datetime import datetime

import bson
from nonebot import logger
from pymongo import UpdateOne

from nonebot_plugin_pixivbot.data.source import MongoDataSource
from nonebot_plugin_pixivbot.model import Illust, User
from .abstract_repo import AbstractPixivRepo
from .lazy_illust import LazyIllust
from .pkg_context import context


@context.register_singleton()
class LocalPixivRepo(AbstractPixivRepo):
    mongo = context.require(MongoDataSource)

    def _make_illusts_cache_loader(self, collection_name: str, arg_name: str, arg: typing.Any, *, skip: int = 0,
                                   limit: int = 0):
        async def cache_loader() -> typing.Optional[typing.List[LazyIllust]]:
            aggregation = [
                {
                    "$match": {arg_name: arg}
                },
                {
                    "$replaceWith": {"illust_id": "$illust_id"}
                },
                {
                    "$unwind": "$illust_id"
                },
            ]

            if skip:
                aggregation.append({"$skip": skip})
            if limit:
                aggregation.append({"$limit": limit})

            aggregation.extend([
                {
                    "$lookup": {
                        "from": "illust_detail_cache",
                        "localField": "illust_id",
                        "foreignField": "illust.id",
                        "as": "illusts"
                    }
                },
                {
                    "$replaceWith": {
                        "$mergeObjects": [
                            "$$ROOT",
                            {"$arrayElemAt": ["$illusts", 0]}
                        ]
                    }
                },
                {
                    "$project": {"_id": 0, "illust": 1, "illust_id": 1}
                }
            ])

            result = self.mongo.db[collection_name].aggregate(aggregation)

            cache = []
            broken = 0
            async for x in result:
                if "illust" in x and x["illust"] is not None:
                    cache.append(LazyIllust(
                        x["illust_id"], Illust.parse_obj(x["illust"])))
                else:
                    cache.append(LazyIllust(x["illust_id"]))
                    broken += 1

            logger.info(
                f"[cache] {len(cache)} got, illust_detail of {broken} are missed")

            if len(cache) != 0:
                return cache
            else:
                return None

        return cache_loader

    def _make_illusts_cache_updater(self, collection_name: str,
                                    arg_name: str,
                                    arg: typing.Any):
        async def cache_updater(content: typing.List[typing.Union[Illust, LazyIllust]]):
            now = datetime.now()
            await self.mongo.db[collection_name].update_one(
                {arg_name: arg},
                {"$set": {
                    "illust_id": [illust.id for illust in content],
                    "update_time": now
                }},
                upsert=True
            )

            opt = []
            for illust in content:
                if isinstance(illust, LazyIllust) and illust.content is not None:
                    illust = illust.content

                if isinstance(illust, Illust):
                    opt.append(UpdateOne(
                        {"illust.id": illust.id},
                        {"$set": {
                            "illust": illust.dict(),
                            "update_time": now
                        }},
                        upsert=True
                    ))
            if len(opt) != 0:
                await self.mongo.db.illust_detail_cache.bulk_write(opt, ordered=False)

        return cache_updater

    async def illust_detail(self, illust_id: int) -> typing.Optional[Illust]:
        cache = await self.mongo.db.illust_detail_cache.find_one({"illust.id": illust_id})
        if cache is not None:
            return Illust.parse_obj(cache["illust"])
        else:
            return None

    async def update_illust_detail(self, illust: Illust):
        await self.mongo.db.illust_detail_cache.update_one(
            {"illust.id": illust.id},
            {"$set": {
                "illust": illust.dict(),
                "update_time": datetime.now()
            }},
            upsert=True
        )

    async def user_detail(self, user_id: int) -> typing.Optional[User]:
        cache = await self.mongo.db.user_detail_cache.find_one({"user.id": user_id})
        if cache is not None:
            return User.parse_obj(cache["user"])
        else:
            return None

    async def update_user_detail(self, user: User):
        await self.mongo.db.user_detail_cache.update_one(
            {"user.id": user.id},
            {"$set": {
                "user": user.dict(),
                "update_time": datetime.now()
            }},
            upsert=True
        )

    def search_illust(self, word: str, *, skip: int = 0, limit: int = 0):
        return self._make_illusts_cache_loader("search_illust_cache", "word", word, skip=skip, limit=limit)()

    def update_search_illust(self, word: str, content: typing.List[typing.Union[Illust, LazyIllust]]):
        return self._make_illusts_cache_updater("search_illust_cache", "word", word)(content)

    async def search_user(self, word: str, *, skip: int = 0, limit: int = 0) -> typing.Optional[typing.List[User]]:
        aggregation = [
            {
                "$match": {"word": word}
            },
            {
                "$replaceWith": {"user_id": "$user_id"}
            },
            {
                "$unwind": "$user_id"
            },
        ]

        if skip:
            aggregation.append({"$skip": skip})
        if limit:
            aggregation.append({"$limit": limit})

        aggregation.extend([
            {
                "$lookup": {
                    "from": "user_detail_cache",
                    "localField": "user_id",
                    "foreignField": "user.id",
                    "as": "users"
                }
            },
            {
                "$replaceWith": {
                    "$mergeObjects": [
                        "$$ROOT",
                        {"$arrayElemAt": ["$users", 0]}
                    ]
                }
            },
            {
                "$project": {"_id": 0, "user": 1, "user_id": 1}
            }
        ])

        result = self.mongo.db.search_user_cache.aggregate(aggregation)

        users = []
        async for x in result:
            if "user" in x and x["user"] is not None:
                users.append(User.parse_obj(x["user"]))
            else:
                users.append(User(id=x["user_id"], name="", account=""))

        if len(users) != 0:
            return users
        else:
            return None

    async def update_search_user(self, word: str, content: typing.List[User]):
        now = datetime.now()
        await self.mongo.db.search_user_cache.update_one(
            {"word": word},
            {"$set": {
                "user_id": [user.id for user in content],
                "update_time": now
            }},
            upsert=True
        )

        opt = []
        for user in content:
            opt.append(UpdateOne(
                {"user.id": user.id},
                {"$set": {
                    "user": user.dict(),
                    "update_time": now
                }},
                upsert=True
            ))
        if len(opt) != 0:
            await self.mongo.db.user_detail_cache.bulk_write(opt, ordered=False)

    def user_illusts(self, user_id: int, *, skip: int = 0, limit: int = 0):
        return self._make_illusts_cache_loader("user_illusts_cache", "user_id", user_id, skip=skip, limit=limit)()

    def update_user_illusts(self, user_id: int, content: typing.List[typing.Union[Illust, LazyIllust]]):
        return self._make_illusts_cache_updater("user_illusts_cache", "user_id", user_id)(content)

    def user_bookmarks(self, user_id: int, *, skip: int = 0, limit: int = 0):
        return self._make_illusts_cache_loader("user_bookmarks_cache", "user_id", user_id, skip=skip, limit=limit)()

    def update_user_bookmarks(self, user_id: int, content: typing.List[typing.Union[Illust, LazyIllust]]):
        return self._make_illusts_cache_updater("user_bookmarks_cache", "user_id", user_id)(content)

    def recommended_illusts(self, *, skip: int = 0, limit: int = 0):
        return self._make_illusts_cache_loader("other_cache", "type", "recommended_illusts", skip=skip, limit=limit)()

    def update_recommended_illusts(self, content: typing.List[typing.Union[Illust, LazyIllust]]):
        return self._make_illusts_cache_updater("other_cache", "type", "recommended_illusts")(content)

    def related_illusts(self, illust_id: int, *, skip: int = 0, limit: int = 0):
        return self._make_illusts_cache_loader("related_illusts_cache", "original_illust_id", illust_id, skip=skip,
                                               limit=limit)()

    def update_related_illusts(self, illust_id: int, content: typing.List[typing.Union[Illust, LazyIllust]]):
        return self._make_illusts_cache_updater("related_illusts_cache", "original_illust_id", illust_id)(content)

    def illust_ranking(self, mode: str, *, skip: int = 0, limit: int = 0):
        return self._make_illusts_cache_loader("other_cache", "type", mode + "_ranking", skip=skip, limit=limit)()

    def update_illust_ranking(self, mode: str, content: typing.List[typing.Union[Illust, LazyIllust]]):
        return self._make_illusts_cache_updater("other_cache", "type", mode + "_ranking")(content)

    async def image(self, illust: Illust) -> typing.Optional[bytes]:
        cache = await self.mongo.db.download_cache.find_one({"illust_id": illust.id})
        if cache is not None:
            return cache["content"]
        else:
            return None

    async def update_image(self, illust: Illust, content: bytes):
        now = datetime.now()
        await self.mongo.db.download_cache.update_one(
            {"illust_id": illust.id},
            {"$set": {
                "content": bson.Binary(content),
                "update_time": now
            }},
            upsert=True
        )

    async def invalidate_cache(self):
        await self.mongo.db.download_cache.delete_many({})
        await self.mongo.db.illust_detail_cache.delete_many({})
        await self.mongo.db.user_detail_cache.delete_many({})
        await self.mongo.db.illust_ranking_cache.delete_many({})
        await self.mongo.db.search_illust_cache.delete_many({})
        await self.mongo.db.search_user_cache.delete_many({})
        await self.mongo.db.user_illusts_cache.delete_many({})
        await self.mongo.db.user_bookmarks_cache.delete_many({})
        await self.mongo.db.other_cache.delete_many({})



from typing import Callable, TypeVar, Generic

from nonebot import logger

from nonebot_plugin_pixivbot.global_context import context as context
from nonebot_plugin_pixivbot.handler.interceptor.interceptor import Interceptor
from nonebot_plugin_pixivbot.postman import PostDestination, Postman
from nonebot_plugin_pixivbot.utils.errors import BadRequestError, QueryError

UID = TypeVar("UID")
GID = TypeVar("GID")


@context.register_singleton()
class DefaultErrorInterceptor(Interceptor[UID, GID], Generic[UID, GID]):
    postman = context.require(Postman)

    async def intercept(self, wrapped_func: Callable,
                        post_dest: PostDestination[UID, GID],
                        silently: bool,
                        **kwargs):
        try:
            await wrapped_func(post_dest=post_dest, silently=silently, **kwargs)
        except TimeoutError:
            logger.warning("Timeout")
            if not silently:
                await self.postman.send_plain_text(f"下载超时", post_dest=post_dest)
        except BadRequestError as e:
            if not silently:
                await self.postman.send_plain_text(str(e), post_dest=post_dest)
        except QueryError as e:
            if not silently:
                await self.postman.send_plain_text(str(e), post_dest=post_dest)
        except Exception as e:
            logger.exception(e)
            if not silently:
                await self.postman.send_plain_text(f"内部错误：{type(e)}{e}", post_dest=post_dest)



from threading import Lock
from typing import TypeVar, Type, Callable

from nonebot.log import logger

T = TypeVar("T")


class Context:
    def __init__(self, parent=None):
        self._parent = parent
        self._container = {}
        self._lazy_container = {}
        self._binding = {}
        self._lock = Lock()

    @property
    def parent(self):
        return self._parent

    @property
    def root(self):
        if self._parent is None:
            return self
        else:
            return self._parent.root

    def register(self, key: Type[T], bean: T):
        """
        register a bean
        """
        self._container[key] = bean
        logger.info(f"registered a bean of type {key}")

    def register_lazy(self, key: Type[T], bean_initializer: Callable[[], T]):
        """
        register a bean lazily
        """
        if key in self._container:
            del self._container[key]
        self._lazy_container[key] = bean_initializer
        logger.info(f"lazily registered a bean of type {key}")

    def register_singleton(self, *args, **kwargs):
        """
        decorator for a class to register a bean lazily
        """

        def decorator(cls):
            self.register_lazy(cls, lambda: cls(*args, **kwargs))
            return cls

        return decorator

    def unregister(self, key: Type[T]):
        """
        unregister the bean of key
        """
        if key in self._container:
            del self._container[key]
        if key in self._lazy_container:
            del self._container[key]

    # 并不好用，因为经常被优化掉import
    # def register_factory(self, key: Type[T]):
    #     """
    #     decorator for a factory method to register a bean lazily
    #     """
    #
    #     def decorator(func: Callable[[], T]):
    #         self.register_lazy(key, func)
    #         return func
    #
    #     return decorator

    def bind(self, key, src_key):
        self._binding[key] = src_key
        logger.info(f"bind bean type {key} to {src_key}")

    def require(self, key: Type[T]) -> T:
        if key in self._binding:
            return self.require(self._binding[key])
        elif key in self._container:
            return self._container[key]
        elif key in self._lazy_container:
            if key in self._lazy_container:
                self.register(key, self._lazy_container[key]())
                del self._lazy_container[key]
            return self._container[key]
        elif self._parent is not None:
            return self._parent.require(key)
        else:
            raise KeyError(key)


__all__ = ("Context",)



import typing

from nonebot import get_driver
from pydantic import BaseSettings, validator
from pydantic.fields import ModelField

from nonebot_plugin_pixivbot.config.block_action import BlockAction
from nonebot_plugin_pixivbot.global_context import context as context


@context.register_singleton(**get_driver().config.dict())
class Config(BaseSettings):
    pixiv_refresh_token: str
    pixiv_mongo_conn_url: str
    pixiv_mongo_database_name: str
    pixiv_proxy: typing.Optional[str]
    pixiv_query_timeout: int = 60
    pixiv_simultaneous_query: int = 8

    pixiv_download_cache_expires_in = 3600 * 24 * 7
    pixiv_illust_detail_cache_expires_in = 3600 * 24 * 7
    pixiv_user_detail_cache_expires_in = 3600 * 24 * 7
    pixiv_illust_ranking_cache_expires_in = 3600 * 6
    pixiv_search_illust_cache_expires_in = 3600 * 24
    pixiv_search_user_cache_expires_in = 3600 * 24
    pixiv_user_illusts_cache_expires_in = 3600 * 24
    pixiv_user_bookmarks_cache_expires_in = 3600 * 24
    pixiv_related_illusts_cache_expires_in = 3600 * 24
    pixiv_other_cache_expires_in = 3600 * 6

    pixiv_block_tags: typing.List[str] = []
    pixiv_block_action: BlockAction = BlockAction.no_image

    pixiv_download_quantity: str = "original"
    pixiv_download_custom_domain: typing.Optional[str]

    @validator('pixiv_download_quantity')
    def download_quantity_validator(cls, v, field: ModelField):
        if v not in ['original', 'square_medium', 'medium', 'large']:
            raise ValueError(
                f'pixiv_compression_enabled is True but {field.name} got None.')
        return v

    pixiv_compression_enabled: bool = False
    pixiv_compression_max_size: typing.Optional[int]
    pixiv_compression_quantity: typing.Optional[float]

    @validator('pixiv_compression_max_size', 'pixiv_compression_quantity')
    def compression_validator(cls, v, values, field: ModelField):
        if values['pixiv_compression_enabled'] and v is None:
            raise ValueError(
                f'pixiv_compression_enabled is True but {field.name} got None.')
        return v

    # pixiv_illust_query_permission = {
    #     "group": [491959457],
    #     "friend": "all"
    # }
    #
    # @validator('pixiv_illust_query_permission')
    # def query_permission_validator(cls, v, field: ModelField):
    #     if v is not None and not isinstance(v, dict):
    #         raise ValueError(f'{field} expected a dict, but got a {type(v)}.')
    #     if "group" in v:
    #         if isinstance(v["group"], list):
    #             for i, x in enumerate(v["group"]):
    #                 if not isinstance(x, int):
    #                     raise ValueError(f'{field}["group"][{i}] expected a int, but got a {type(x)}.')
    #         elif v["group"] != "all":
    #             raise ValueError(f'{field}["group"] expected "all" or a list, but got a {type(v["group"])}.')
    #     if "friend" in v:
    #         if isinstance(v["friend"], list):
    #             for i, x in enumerate(v["group"]):
    #                 if not isinstance(x, int):
    #                     raise ValueError(f'{field}["friend"][{i}] expected a int, but got a {type(x)}.')
    #         elif v["friend"] != "all":
    #             raise ValueError(f'{field}["friend"] expected "all" or a list, but got a {type(v["friend"])}.')

    pixiv_query_cooldown = 0
    pixiv_no_query_cooldown_users = []
    pixiv_max_item_per_query = 10

    pixiv_tag_translation_enabled = True

    pixiv_more_enabled = True
    pixiv_query_expires_in = 10 * 60

    pixiv_illust_query_enabled = True

    pixiv_ranking_query_enabled = True
    pixiv_ranking_default_mode = "day"
    pixiv_ranking_default_range = [1, 3]
    pixiv_ranking_fetch_item = 150
    pixiv_ranking_max_item_per_query = 10

    @validator('pixiv_ranking_default_mode')
    def ranking_default_mode_validator(cls, v, field: ModelField):
        if v not in ['day', 'week', 'month', 'day_male', 'day_female', 'week_original', 'week_rookie', 'day_manga']:
            raise ValueError(f'illegal {field.name} value: {v}')
        return v

    @validator('pixiv_ranking_default_range')
    def ranking_default_range_validator(cls, v, field: ModelField):
        if len(v) < 2 or v[0] > v[1]:
            raise ValueError(f'illegal {field.name} value: {v}')
        return v

    pixiv_random_illust_query_enabled = True
    pixiv_random_illust_method = "bookmark_proportion"
    pixiv_random_illust_min_bookmark = 0
    pixiv_random_illust_min_view = 0
    pixiv_random_illust_max_page = 20
    pixiv_random_illust_max_item = 500

    pixiv_random_recommended_illust_query_enabled = True
    pixiv_random_recommended_illust_method = "uniform"
    pixiv_random_recommended_illust_min_bookmark = 0
    pixiv_random_recommended_illust_min_view = 0
    pixiv_random_recommended_illust_max_page = 40
    pixiv_random_recommended_illust_max_item = 1000

    pixiv_random_related_illust_query_enabled = True
    pixiv_random_related_illust_method = "bookmark_proportion"
    pixiv_random_related_illust_min_bookmark = 0
    pixiv_random_related_illust_min_view = 0
    pixiv_random_related_illust_max_page = 4
    pixiv_random_related_illust_max_item = 100

    pixiv_random_user_illust_query_enabled = True
    pixiv_random_user_illust_method = "timedelta_proportion"
    pixiv_random_user_illust_min_bookmark = 0
    pixiv_random_user_illust_min_view = 0
    pixiv_random_user_illust_max_page = 2 ** 31
    pixiv_random_user_illust_max_item = 2 ** 31

    pixiv_random_bookmark_query_enabled = True
    pixiv_random_bookmark_user_id: typing.Optional[int] = None
    pixiv_random_bookmark_method = "uniform"
    pixiv_random_bookmark_min_bookmark = 0
    pixiv_random_bookmark_min_view = 0
    pixiv_random_bookmark_max_page = 2 ** 31
    pixiv_random_bookmark_max_item = 2 ** 31

    @validator('pixiv_random_illust_method', 'pixiv_random_recommended_illust_method',
               'pixiv_random_user_illust_method', 'pixiv_random_bookmark_method')
    def random_method_validator(cls, v, field: ModelField):
        if v not in ['bookmark_proportion', 'view_proportion', 'timedelta_proportion', 'uniform']:
            raise ValueError(f'illegal {field.name} value: {v}')
        return v

    pixiv_poke_action: typing.Optional[str] = "random_recommended_illust"

    @validator('pixiv_poke_action')
    def pixiv_poke_action_validator(cls, v, field: ModelField):
        if v not in [None, "", "ranking", "random_recommended_illust", "random_bookmark"]:
            raise ValueError(f'illegal {field.name} value: {v}')
        return v

    class Config:
        extra = "ignore"


__all__ = ("Config",)



from asyncio import create_task, wait
from typing import Optional, Sequence

from nonebot_plugin_pixivbot.global_context import context as context
from nonebot_plugin_pixivbot.model import Illust
from nonebot_plugin_pixivbot.postman.model.illust_message import IllustMessageModel
from nonebot_plugin_pixivbot.postman.model.illust_messages import IllustMessagesModel
from nonebot_plugin_pixivbot.postman.post_destination import PostDestination
from nonebot_plugin_pixivbot.postman.postman import Postman


async def post_illust(illust: Illust, *,
                      header: Optional[str] = None,
                      number: Optional[int] = None,
                      post_dest: PostDestination):
    postman = context.require(Postman)
    model = await IllustMessageModel.from_illust(illust, header=header, number=number)
    if model is not None:
        await postman.send_illust(model, post_dest=post_dest)


async def post_illusts(illusts: Sequence[Illust], *,
                       header: Optional[str] = None,
                       number: Optional[int] = None,
                       post_dest: PostDestination):
    postman = context.require(Postman)
    tasks = [
        create_task(
            IllustMessageModel.from_illust(x, number=number + i if number is not None else None)
        ) for i, x in enumerate(illusts)
    ]
    await wait(tasks)

    messages = []
    for t in tasks:
        result = await t
        if result:
            messages.append(result)

    model = IllustMessagesModel(header=header, messages=messages)
    await postman.send_illusts(model, post_dest=post_dest)


__all__ = ("post_illust", "post_illusts")



from typing import Generic, TypeVar, Sequence, Any

from nonebot_plugin_pixivbot.global_context import context as context
from nonebot_plugin_pixivbot.handler.common.common_handler import CommonHandler
from nonebot_plugin_pixivbot.postman import PostDestination, post_illusts
from nonebot_plugin_pixivbot.service.pixiv_account_binder import PixivAccountBinder
from nonebot_plugin_pixivbot.utils.errors import BadRequestError

UID = TypeVar("UID")
GID = TypeVar("GID")


@context.root.register_singleton()
class RandomBookmarkHandler(CommonHandler[UID, GID], Generic[UID, GID]):
    binder = context.require(PixivAccountBinder)

    @classmethod
    def type(cls) -> str:
        return "random_bookmark"

    @classmethod
    def enabled(cls) -> bool:
        return cls.conf.pixiv_random_bookmark_query_enabled

    def parse_args(self, args: Sequence[Any], post_dest: PostDestination[UID, GID]) -> dict:
        pixiv_user_id = 0
        sender_user_id = post_dest.user_id

        if len(args) > 0:
            try:
                pixiv_user_id = int(args[0])
            except ValueError:
                raise BadRequestError(f"{args[0]}不是合法的ID")

        return {"pixiv_user_id": pixiv_user_id, "sender_user_id": sender_user_id}

    async def actual_handle(self, *, sender_user_id: UID,
                            pixiv_user_id: int = 0,
                            count: int = 1,
                            post_dest: PostDestination[UID, GID],
                            silently: bool = False):
        if not pixiv_user_id and sender_user_id:
            pixiv_user_id = await self.binder.get_binding(sender_user_id)

        if not pixiv_user_id:
            pixiv_user_id = self.conf.pixiv_random_bookmark_user_id

        if not pixiv_user_id:
            raise BadRequestError("无效的Pixiv账号，或未绑定Pixiv账号")

        illusts = await self.service.random_bookmark(pixiv_user_id, count=count)

        # 记录请求
        self.record_req(pixiv_user_id, post_dest=post_dest, count=count)
        # 记录结果
        if len(illusts) == 1:
            self.record_resp_illust(illusts[0].id, post_dest=post_dest)

        await post_illusts(illusts,
                           header="这是您点的私家车",
                           post_dest=post_dest)



from abc import ABC
from typing import TypeVar, Generic

from nonebot_plugin_pixivbot.global_context import context as context
from nonebot_plugin_pixivbot.handler.common.recorder import Req, Recorder
from nonebot_plugin_pixivbot.handler.handler import Handler
from nonebot_plugin_pixivbot.handler.interceptor.combined_interceptor import CombinedInterceptor
from nonebot_plugin_pixivbot.handler.interceptor.cooldown_interceptor import CooldownInterceptor
from nonebot_plugin_pixivbot.handler.interceptor.default_error_interceptor import DefaultErrorInterceptor
from nonebot_plugin_pixivbot.model import PostIdentifier
from nonebot_plugin_pixivbot.postman import PostDestination
from nonebot_plugin_pixivbot.service.pixiv_service import PixivService

UID = TypeVar("UID")
GID = TypeVar("GID")


class CommonHandler(Handler[UID, GID], ABC, Generic[UID, GID]):
    service = context.require(PixivService)
    recorder = context.require(Recorder)

    interceptor = CombinedInterceptor.from_iterable(
        [
            context.require(DefaultErrorInterceptor),
            context.require(CooldownInterceptor)
        ]
    )

    def record_req(self, *args,
                   post_dest: PostDestination[UID, GID],
                   **kwargs):
        self.recorder.record_req(Req(self, *args, **kwargs), PostIdentifier.from_post_dest(post_dest))

    def record_resp_illust(self, illust_id: int,
                           post_dest: PostDestination[UID, GID]):
        self.recorder.record_resp(illust_id, PostIdentifier.from_post_dest(post_dest))



from typing import Optional, Sequence

from pydantic import BaseModel

from nonebot_plugin_pixivbot.postman.model.illust_message import IllustMessageModel


class IllustMessagesModel(BaseModel):
    header: Optional[str] = None
    messages: Sequence[IllustMessageModel]

    def flat_first(self) -> IllustMessageModel:
        return self.messages[0].copy(update={"header": self.header})



from typing import TypeVar, Generic, Sequence, Any

from nonebot_plugin_pixivbot.global_context import context as context
from nonebot_plugin_pixivbot.handler.common.common_handler import CommonHandler
from nonebot_plugin_pixivbot.postman import PostDestination, post_illusts

UID = TypeVar("UID")
GID = TypeVar("GID")


@context.root.register_singleton()
class RandomIllustHandler(CommonHandler, Generic[UID, GID]):
    @classmethod
    def type(cls) -> str:
        return "random_illust"

    @classmethod
    def enabled(cls) -> bool:
        return cls.conf.pixiv_random_illust_query_enabled

    def parse_args(self, args: Sequence[Any], post_dest: PostDestination[UID, GID]) -> dict:
        return {"word": args[0]}

    async def actual_handle(self, *, word: str,
                            count: int = 1,
                            post_dest: PostDestination[UID, GID],
                            silently: bool = False):
        illusts = await self.service.random_illust(word, count=count)

        # 记录请求
        self.record_req(word, post_dest=post_dest, count=count)
        # 记录结果
        if len(illusts) == 1:
            self.record_resp_illust(illusts[0].id, post_dest=post_dest)

        await post_illusts(illusts,
                           header=f"这是您点的{word}图",
                           post_dest=post_dest)



from typing import TypeVar, Optional, Generic

from nonebot_plugin_pixivbot.data.source import MongoDataSource
from nonebot_plugin_pixivbot.global_context import context as context
from nonebot_plugin_pixivbot.model.pixiv_binding import PixivBinding

UID = TypeVar("UID")


@context.register_singleton()
class PixivBindingRepo(Generic[UID]):
    mongo = context.require(MongoDataSource)

    async def get(self, adapter: str, user_id: UID) -> Optional[PixivBinding]:
        return await self.mongo.db.pixiv_binding.find_one(
            {"adapter": adapter, "user_id": user_id}
        )

    async def update(self, binding: PixivBinding):
        await self.mongo.db.pixiv_binding.update_one(
            {"adapter": binding.adapter, "user_id": binding.user_id},
            {"$set": {
                "pixiv_user_id": binding.pixiv_user_id
            }},
            upsert=True
        )

    async def remove(self, adapter: str, user_id: UID) -> bool:
        cnt = await self.mongo.db.pixiv_binding.delete_one(
            {"adapter": adapter, "user_id": user_id}
        )
        return cnt == 1


__all__ = ("PixivBindingRepo",)



from .illust_handler import IllustHandler
from .more_handler import MoreHandler
from .random_bookmark_handler import RandomBookmarkHandler
from .random_illust_handler import RandomIllustHandler
from .random_recommended_illust_handler import RandomRecommendedIllustHandler
from .random_related_illust_handler import RandomRelatedIllustHandler
from .random_user_illust_handler import RandomUserIllustHandler
from .ranking_handler import RankingHandler

__all__ = ("RandomBookmarkHandler", "RandomRecommendedIllustHandler",
           "RankingHandler", "RandomIllustHandler",
           "RandomUserIllustHandler", "IllustHandler",
           "MoreHandler", "RandomRelatedIllustHandler")



from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.schedulers.base import BaseScheduler
from nonebot import require

from nonebot_plugin_pixivbot.utils.context import Context


def asyncio_scheduler_provider(context: Context):
    context.register_lazy(AsyncIOScheduler, lambda: require("nonebot_plugin_apscheduler").scheduler)


def base_scheduler_provider(context: Context):
    context.bind(BaseScheduler, AsyncIOScheduler)


providers = [asyncio_scheduler_provider, base_scheduler_provider]



from datetime import datetime
from math import ceil
from typing import TypeVar, Generic, Callable

from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.global_context import context as context
from nonebot_plugin_pixivbot.handler.interceptor.interceptor import Interceptor
from nonebot_plugin_pixivbot.postman import PostDestination, Postman

UID = TypeVar("UID")
GID = TypeVar("GID")


@context.register_singleton()
class CooldownInterceptor(Interceptor[UID, GID], Generic[UID, GID]):
    conf = context.require(Config)
    postman = context.require(Postman)

    def __init__(self):
        self.last_query_time = dict[UID, datetime]()

    def get_cooldown(self, user_id: UID) -> int:
        if self.conf.pixiv_query_cooldown == 0 or user_id in self.conf.pixiv_no_query_cooldown_users:
            return 0

        now = datetime.now()
        if user_id not in self.last_query_time:
            self.last_query_time[user_id] = now
            return 0
        else:
            delta = now - self.last_query_time[user_id]
            cooldown = self.conf.pixiv_query_cooldown - ceil(delta.total_seconds())
            if cooldown > 0:
                return cooldown
            else:
                self.last_query_time[user_id] = now
                return 0

    async def intercept(self, wrapped_func: Callable,
                        post_dest: PostDestination[UID, GID],
                        silently: bool,
                        **kwargs):
        cooldown = self.get_cooldown(post_dest.user_id)
        if cooldown:
            if not silently:
                await self.postman.send_plain_text(f"你的CD还有{cooldown}s转好", post_dest=post_dest)
        else:
            await wrapped_func(post_dest=post_dest, silently=silently, **kwargs)


__all__ = ("CooldownInterceptor",)



from typing import TypeVar, Generic, Union, Sequence, Any

from nonebot_plugin_pixivbot.global_context import context as context
from nonebot_plugin_pixivbot.handler.common.common_handler import CommonHandler
from nonebot_plugin_pixivbot.postman import PostDestination, post_illusts

UID = TypeVar("UID")
GID = TypeVar("GID")


@context.root.register_singleton()
class RandomUserIllustHandler(CommonHandler[UID, GID], Generic[UID, GID]):
    @classmethod
    def type(cls) -> str:
        return "random_user_illust"

    @classmethod
    def enabled(cls) -> bool:
        return cls.conf.pixiv_random_user_illust_query_enabled

    async def parse_args(self, args: Sequence[Any], post_dest: PostDestination[UID, GID]) -> dict:
        user = args[0]
        if isinstance(user, str):
            user = await self.service.get_user(user)

        return {"user": user.id}

    async def actual_handle(self, *, user: Union[str, int],
                            count: int = 1,
                            post_dest: PostDestination[UID, GID],
                            silently: bool = False):
        userinfo, illusts = await self.service.random_user_illust(user, count=count)

        # 记录请求
        self.record_req(userinfo.id, post_dest=post_dest, count=count)
        # 记录结果
        if len(illusts) == 1:
            self.record_resp_illust(illusts[0].id, post_dest=post_dest)

        await post_illusts(illusts,
                           header=f"这是您点的{userinfo.name}老师({userinfo.id})的图",
                           post_dest=post_dest)



class QueryError(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class BadRequestError(Exception):
    def __init__(self, message=None):
        super().__init__()
        self.message = message

    def __str__(self):
        return self.message


__all__ = ("QueryError", "BadRequestError")



from .post_destination import PostDestination, PostDestinationFactory
from .post_helper import post_illust, post_illusts
from .postman import Postman

__all__ = ("Postman", "PostDestination", "PostDestinationFactory",
           "post_illust", "post_illusts")



from typing import List, Sequence, Union, Tuple

from nonebot import logger

from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.data.local_tag_repo import LocalTagRepo
from nonebot_plugin_pixivbot.data.pixiv import LazyIllust, PixivRepo
from nonebot_plugin_pixivbot.global_context import context as context
from nonebot_plugin_pixivbot.model import Illust, User
from nonebot_plugin_pixivbot.utils.errors import BadRequestError, QueryError
from .roulette import roulette


@context.register_singleton()
class PixivService:
    conf = context.require(Config)
    data_source = context.require(PixivRepo)
    local_tags = context.require(LocalTagRepo)

    async def _choice_and_load(self, illusts: List[LazyIllust], random_method: str, count: int) -> List[Illust]:
        if count <= 0:
            raise BadRequestError("不合法的请求数量")
        if count > self.conf.pixiv_max_item_per_query:
            raise BadRequestError("数量超过单次请求上限")
        if count > len(illusts):
            raise QueryError("别看了，没有的。")

        winners = roulette(illusts, random_method, count)
        logger.info(f"choice {[x.id for x in winners]}")
        return [await x.get() for x in winners]

    async def illust_ranking(self, mode: str, range: Sequence[int]) -> List[Illust]:
        start, end = range
        illusts = await self.data_source.illust_ranking(mode, skip=start - 1, limit=end - start + 1)

        return [await x.get() for x in illusts]

    async def illust_detail(self, illust: int) -> Illust:
        return await self.data_source.illust_detail(illust)

    async def random_illust(self, word: str, *, count: int = 1) -> List[Illust]:
        if self.conf.pixiv_tag_translation_enabled:
            # 只有原word不是标签时获取翻译（例子：唐可可）
            tag = await self.local_tags.get_by_name(word)
            if not tag:
                tag = await self.local_tags.get_by_translated_name(word)
                if tag:
                    logger.info(f"found translation {word} -> {tag.name}")
                    word = tag.name

        illusts = await self.data_source.search_illust(word)
        return await self._choice_and_load(illusts, self.conf.pixiv_random_illust_method, count)

    async def get_user(self, user: Union[str, int]) -> User:
        if isinstance(user, str):
            users = await self.data_source.search_user(user)
            if len(users) == 0:
                raise QueryError("未找到用户")
            else:
                return users[0]
        else:
            return await self.data_source.user_detail(user)

    async def random_user_illust(self, user: Union[str, int], *, count: int = 1) -> Tuple[User, List[Illust]]:
        user = await self.get_user(user)
        illusts = await self.data_source.user_illusts(user.id)
        illust = await self._choice_and_load(illusts, self.conf.pixiv_random_user_illust_method, count)
        return user, illust

    async def random_recommended_illust(self, *, count: int = 1) -> List[Illust]:
        illusts = await self.data_source.recommended_illusts()
        return await self._choice_and_load(illusts, self.conf.pixiv_random_recommended_illust_method, count)

    async def random_bookmark(self, pixiv_user_id: int = 0, *, count: int = 1) -> List[Illust]:
        illusts = await self.data_source.user_bookmarks(pixiv_user_id)
        return await self._choice_and_load(illusts, self.conf.pixiv_random_bookmark_method, count)

    async def random_related_illust(self, illust_id: int, *, count: int = 1) -> List[Illust]:
        if illust_id == 0:
            raise BadRequestError("你还没有发送过请求")

        illusts = await self.data_source.related_illusts(illust_id)
        return await self._choice_and_load(illusts, self.conf.pixiv_random_related_illust_method, count)


__all__ = ("PixivService",)



import typing

from pymongo import *

from nonebot_plugin_pixivbot.data.source import MongoDataSource
from nonebot_plugin_pixivbot.global_context import context as context
from nonebot_plugin_pixivbot.model import Tag


@context.register_singleton()
class LocalTagRepo:
    mongo = context.require(MongoDataSource)

    async def insert(self, tag: Tag) -> typing.NoReturn:
        await self.mongo.db.local_tags.updateOne(
            {"name": tag.name},
            {"$setOnInsert": {"translated_name": tag.translated_name}},
            upsert=True
        )

    async def insert_many(self, tags: typing.Iterable[Tag]) -> typing.NoReturn:
        opt = []

        for tag in tags:
            opt.append(UpdateOne(
                {"name": tag.name},
                {"$setOnInsert": {"translated_name": tag.translated_name}},
                upsert=True
            ))

        if len(opt) != 0:
            await self.mongo.db.local_tags.bulk_write(opt, ordered=False)

    async def get_by_name(self, name: str) -> typing.Optional[Tag]:
        result = await self.mongo.db.local_tags.find_one({"name": name})
        if result:
            return Tag.parse_obj(result)
        else:
            return None

    async def get_by_translated_name(self, translated_name: str) -> typing.Optional[Tag]:
        result = await self.mongo.db.local_tags.find_one({"translated_name": translated_name})
        if result:
            return Tag.parse_obj(result)
        else:
            return None


__all__ = ("LocalTagRepo",)
